# Third-party
import numpy as np
from astropy import units as u
from astropy.modeling.models import Sersic2D

# Project
from ..util import check_random_state, check_units, check_xy_dim
from ..log import logger


__all__ = ['xy_from_grid', 'sersic_xy', 'plummer_xy', 'uniform_xy']


def _check_offset(xy_dim, dx, dy):
    """Check if center of source falls within image."""
    center = np.array([xy_dim[0]//2 + dx, xy_dim[1]//2 + dy])
    if np.any(center <= 0.) or np.any(center >= xy_dim):
        raise Exception(f'Center = {center} falls outside of the image.')


def xy_from_grid(num_stars, model, xy_dim, sample_dim=None,
                 dx=0, dy=0, drop_outside=False, random_state=None):
    """
    Sample xy positions from a discrete grid weighted by an arbitrary model.

    Parameters
    ----------
    num_stars : int
        Number of stars (i.e., positions) to sample.
    model : `~astropy.modeling.Fittable2DModel`
        A two-dimensional ``astropy`` model or a callable object that takes
        x and y grid positions arguments: `model(xx, yy)`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    sample_dim : int or None, optional
        Square dimension to sample within. This is useful if the mock image
        size is small compared to the scale of the model. To be useful,
        `sample_dim` should be larger than `xy_dim`.
    dx : float, optional
        Shift from center of image in pixels in the x direction.
    dy : float, optional
        Shift from center of image in pixels in the y direction.
    drop_outside : bool, optional
        If True, drop all stars that fall outside of the image. In this case,
        the returned `~numpy.ndarray` will not be masked.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.

    Returns
    -------
    xy : `~numpy.ma.MaskedArray` or `~numpy.ndarray`
        Masked numpy array of xy positions. Positions that fall outside the
        mock image are masked. If ``drop_outside == True``, then the stars
        that fall outside of the image will be dropped and the returned array
        will not be masked.
    """
    xy_dim = check_xy_dim(xy_dim)
    _check_offset(xy_dim, dx, dy)

    if sample_dim is None:
        sample_dim = max(xy_dim)

    num_pix = 2 * sample_dim**2
    byte_per_pixel = 8
    memory_in_GB = (byte_per_pixel * num_pix) / 1e9
    if memory_in_GB > 1:
        msg = "This sampling grid will exceed 1 Gb in memory usage."
        logger.warning(msg)

    x_0, y_0 = sample_dim//2, sample_dim//2
    yy, xx = np.meshgrid(np.arange(sample_dim),
                         np.arange(sample_dim))

    image = model(xx, yy)
    weights = image.flatten()
    weights /= np.sum(weights)

    rng = check_random_state(random_state)
    samples = rng.choice(len(weights), size=int(num_stars), p=weights)
    x, y = np.unravel_index(samples, (sample_dim, sample_dim), order='C')

    shift = (sample_dim - xy_dim) // 2
    x -= shift[0]
    y -= shift[1]
    x_0 -= shift[0]
    y_0 -= shift[1]

    x += dx
    y += dy
    xy = np.vstack([x, y]).T
    xy = np.ma.masked_array(xy, mask=np.zeros_like(xy, dtype=bool))

    if xy_dim[0] < sample_dim or xy_dim[1] < sample_dim:
        outside_image = x < 0
        outside_image |= x > xy_dim[0] - 1
        outside_image |= y < 0
        outside_image |= y > xy_dim[1] - 1

        if outside_image.sum() > 0:
            msg = '{} stars outside the image'.format(outside_image.sum())
            logger.warning(msg)
            xy.mask = np.column_stack((outside_image, outside_image))

            if drop_outside:
                xy = xy[~outside_image].data

    return xy


def sersic_xy(num_stars, distance, xy_dim, pixel_scale, r_eff=500*u.pc, n=1.0,
              theta=45*u.deg, ellip=0.3, num_r_eff=10, dx=0, dy=0,
              drop_outside=False, random_state=None):
    """
    Sample xy positions from a two-dimensional Sersic distribution.


    .. note::
        To sample the Sersic distribution, `~artpop.space.sample.xy_from_grid`
        is used. This means that large sources (and/or large `num_r_eff`) eat
        up a lot of memory.

    Parameters
    ----------
    num_stars : int
        Number of stars (i.e., positions) to sample.
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `~astropy.units.Mpc`.
    xy_dim : list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `~astropy.units.arcsec` per `~astropy.units.pixels`.
    r_eff : float or `~astropy.units.Quantity`, optional
        Effective radius of the source. If a float is given, the units are
        assumed to be `~astropy.units.kpc`. Must be greater than zero.
    n : float, optional
        Sersic index. Must be greater than zero.
    theta : float or `~astropy.units.Quantity`, optional
        Rotation angle, counterclockwise from the positive x-axis. If a float
        is given, the units are assumed to be `degree`.
    ellip : float, optional
        Ellipticity defined as `1 - b/a`, where `b` is the semi-minor axis
        and `a` is the semi-major axis.
    num_r_eff : float, optional
        Number of r_eff to sample positions within. This parameter is needed
        because the current Sersic sampling function samples from within a
        discrete grid. Default is 10.
    dx : float, optional
        Shift from center of image in pixels in the x direction.
    dy : float, optional
        Shift from center of image in pixels in the y direction.
    drop_outside : bool, optional
        If True, drop all stars that fall outside of the image. In this case,
        the returned `~numpy.ndarray` will not be masked.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.

    Returns
    -------
    xy : `~numpy.ma.MaskedArray`
        Masked numpy array of xy positions. Positions that fall outside the
        mock image are masked.
    """
    if n <= 0:
        raise Exception('Sersic index n must be greater than zero.')

    xy_dim = check_xy_dim(xy_dim)
    _check_offset(xy_dim, dx, dy)

    r_eff = check_units(r_eff, 'kpc').to('Mpc').value
    theta = check_units(theta, 'deg').to('radian').value
    distance = check_units(distance, 'Mpc').to('Mpc').value
    pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)

    if r_eff <= 0:
        raise Exception('Effective radius must be greater than zero.')

    r_pix = np.arctan2(r_eff, distance) * u.radian.to('arcsec') * u.arcsec
    r_pix = r_pix.to('pixel', u.pixel_scale(pixel_scale)).value
    sample_dim = 2 * np.ceil(r_pix * num_r_eff).astype(int) + 1
    x_0, y_0 = sample_dim//2, sample_dim//2

    model = Sersic2D(x_0=x_0, y_0=y_0, amplitude=1, r_eff=r_pix,
                     n=n, ellip=ellip, theta=theta)

    xy = xy_from_grid(num_stars, model, xy_dim, sample_dim, dx, dy,
                      drop_outside=drop_outside, random_state=random_state)

    return xy


def plummer_xy(num_stars, distance, xy_dim, pixel_scale, scale_radius=10*u.pc,
               dx=0, dy=0, drop_outside=False, random_state=None):
    """
    Sample xy positions from a two-dimensional Plummer distributions using
    inverse transform sampling.

    Parameters
    ----------
    num_stars : int
        Number of stars (i.e., positions) to sample.
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `~astropy.units.Mpc`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `~astropy.units.arcsec` per `~astropy.units.pixels`.
    scale_radius : float or `~astropy.units.Quantity`, optional
        Scale radius of the source. If a float is given, the units are
        assumed to be `~astropy.units.kpc`. Must be greater than zero.
    dx : float, optional
        Shift from center of image in pixels in the x direction.
    dy : float, optional
        Shift from center of image in pixels in the y direction.
    drop_outside : bool, optional
        If True, drop all stars that fall outside of the image. In this case,
        the returned `~numpy.ndarray` will not be masked.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.

    Returns
    -------
    xy : `~numpy.ma.MaskedArray` or `~numpy.ndarray`
        Masked numpy array of xy positions. Positions that fall outside the
        mock image are masked. If ``drop_outside == True``, then the stars
        that fall outside of the image will be dropped and the returned array
        will not be masked.
    """
    xy_dim = check_xy_dim(xy_dim)
    _check_offset(xy_dim, dx, dy)

    scale_radius = check_units(scale_radius, 'kpc').to('Mpc').value
    distance = check_units(distance, 'Mpc').to('Mpc').value
    r_sky = np.arctan2(scale_radius, distance) * u.radian.to('arcsec')
    pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)

    rng = check_random_state(random_state)
    s = rng.random(size=int(num_stars))
    r = u.arcsec * r_sky / np.sqrt(s**(-2 / 3) - 1)
    r = r.to('pixel', u.pixel_scale(pixel_scale)).value

    phi = rng.uniform(0, 2 * np.pi, size=len(r))
    theta = np.arccos(2 * rng.random(size=len(r)) - 1)
    xy = np.zeros((len(r), 2))

    shift = xy_dim /  2
    xy[:, 0] = r * np.cos(phi) * np.sin(theta) + shift[0] + dx
    xy[:, 1] = r * np.sin(phi) * np.sin(theta) + shift[1] + dy

    x_outside = (xy[:, 0] < 0) | (xy[:, 0] >= xy_dim[0])
    y_outside = (xy[:, 1] < 0) | (xy[:, 1] >= xy_dim[1])
    xy = np.ma.masked_array(xy, mask=np.zeros_like(xy, dtype=bool))

    outside_image = x_outside | y_outside
    if outside_image.sum() > 0:
        msg = '{} stars outside the image'.format(outside_image.sum())
        logger.warning(msg)
        xy.mask = np.column_stack((outside_image, outside_image))

        if drop_outside:
            xy = xy[~outside_image].data

    return xy


def uniform_xy(num_stars, xy_dim, random_state=None):
    """
    Sample xy positions from a uniform distribution.

    Parameters
    ----------
    num_stars : int
        Number of stars (i.e., positions) to sample.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.

    Returns
    -------
    xy : `~numpy.array`
        xy pixel positions.
    """
    num_stars = int(num_stars)
    xy_dim = check_xy_dim(xy_dim)
    rng = check_random_state(random_state)
    xy = np.vstack([rng.uniform(0, xy_dim[0], num_stars),
                    rng.uniform(0, xy_dim[1], num_stars)]).T
    return xy



# Standard library
import os
import tarfile

# Third-party
import requests
import numpy as np
from astropy import units as u
from astropy.utils.misc import isiterable

# Project
from . import MIST_PATH
from .log import logger
from .filters import phot_system_list


__all__ = ['check_random_state',
           'check_units',
           'check_odd',
           'check_xy_dim',
           'embed_slices',
           'fetch_mist_grid_if_needed']


def check_random_state(seed):
    """
    Turn seed into a `~numpy.random.RandomState` instance.

    Parameters
    ----------
    seed : `None`, int, list of ints, or `~numpy.random.RandomState`
        If ``seed`` is `None`, return the `~numpy.random.RandomState`
        singleton used by ``numpy.random``.  If ``seed`` is an `int`,
        return a new `~numpy.random.RandomState` instance seeded with
        ``seed``.  If ``seed`` is already a `~numpy.random.RandomState`,
        return it.  Otherwise raise ``ValueError``.

    Returns
    -------
    random_state : `~numpy.random.RandomState`
        RandomState object.

    Notes
    -----
    This routine is adapted from scikit-learn. See
    http://scikit-learn.org/stable/developers/utilities.html#validation-tools.
    """
    import numbers

    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    if type(seed)==list:
        if type(seed[0])==int:
            return np.random.RandomState(seed)

    raise ValueError('{0!r} cannot be used to seed a numpy.random.RandomState'
                     ' instance'.format(seed))


def check_units(value, default_unit):
    """
    Check if an object has units. If not, apply the default unit.

    Parameters
    ----------
    value : float, list-like, or `~astropy.units.Quantity`
        Parameter that has units.
    default_unit : str or `astropy` unit
        The default unit to apply to `value` if it does not have units.

    Returns
    -------
    quantity : `~astropy.units.Quantity`
        `value` with ``astropy`` units.
    """
    t = type(default_unit)
    if type(value) == u.Quantity:
        quantity = value
    elif (t == u.IrreducibleUnit) or (t == u.Unit) or (t == u.CompositeUnit):
        quantity = value * default_unit
    elif t == str:
        quantity = value * getattr(u, default_unit)
    else:
        raise Exception('default_unit must be an astropy unit or string')
    return quantity


def check_odd(val, name='value'):
    """
    Raise Exception if  `val` is not odd.
    """
    if isiterable(val):
        if np.any(np.asarray(val) % 2 == 0):
            raise Exception(f'{name} must be odd')
    else:
        if val % 2 == 0:
            raise Exception(f'{name} must be odd')


def check_xy_dim(xy_dim, force_odd=True):
    """
    Check the format of `xy_dim`.

    Parameters
    ----------
    xy_dim : int or list-like
        The dimensions of mock image in xy units. If `int` is given, it is
        assumed to be both the x and y dimensions and (xy_dim, xy_dim) will be
        returned. Otherwise xy_dim will be returned.
    force_odd : bool, optional
        If True (default), force both the x and y dimensions to be odd.

    Returns
    -------
    xy_dim : `~numpy.ndarray`
        Dimensions of mock image in xy units.
    """
    if not isiterable(xy_dim):
        xy_dim = [xy_dim, xy_dim]
    xy_dim = np.asarray(xy_dim).astype(int)
    if force_odd:
        check_odd(xy_dim, 'xy dimensions')
    return xy_dim


def embed_slices(center, model_shape, image_shape):
    """
    Get slices to embed smaller model array into larger image array.

    Parameters
    ----------
    center : `~numpy.ndarray`
        Center of array in the image coordinates.
    model_shape : tuple
        Shape of the array to embed (dimensions must be odd).
    image_shape : tuple
        Shape of the main image array.

    Returns
    -------
    img_slice, mod_slice : tuples of slices
        Slicing indices. To embed array in image,
        use the following: image[img_slice] = model[mod_slice]
    """
    model_shape = np.asarray(model_shape)
    image_shape = np.asarray(image_shape)

    check_odd(model_shape, 'embed_slices array shape')

    imin = center - model_shape//2
    imax = center + model_shape//2

    amin = (imin < np.array([0,0])) * (-imin)
    amax = model_shape * (imax <= image_shape - 1) +\
           (model_shape - (imax - image_shape + 1)) * (imax > image_shape - 1)

    imin = np.maximum(imin, np.array([0, 0]))
    imax = np.minimum(imax, np.array(image_shape)-1)
    imax += 1

    img_slice = np.s_[imin[0]:imax[0], imin[1]:imax[1]]
    mod_slice = np.s_[amin[0]:amax[0], amin[1]:amax[1]]

    return img_slice, mod_slice


def fetch_mist_grid_if_needed(phot_system, v_over_vcrit=0.4,
                              mist_path=MIST_PATH, overwrite=False):
    """
    If needed, fetch MIST grid from http://waps.cfa.harvard.edu/MIST.

    Parameters
    ----------
    phot_system : str
        Photometric system grid to fetch. Must be a supported ArtPop filter
        system, where are listed in `~artpop.filters.phot_system_list`.
    v_over_vcrit : float, optional
        Rotation rate divided by the critical surface linear velocity. Current
        options are 0.4 (default) and 0.0.
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the default location of ~/.artpop/mist (or the `MIST_PATH`
        environment variable if you have it set).
    overwrite : bool, optional
        If True, force an overwrite of grid if it exists.
     """
    if phot_system not in phot_system_list:
        raise Exception(f'Photometric system must be in {phot_system_list}.')
    version = 1.2
    url = f'http://waps.cfa.harvard.edu/MIST/data/tarballs_v{version}'
    url += f'/MIST_v{version}_vvcrit{v_over_vcrit}_{phot_system}.txz'
    tarball = os.path.join(mist_path, os.path.basename(url))
    grid_path = tarball.replace('.txz', '')
    if overwrite or not os.path.isdir(grid_path):
        logger.info(f'Fetching MIST synthetic photometry grid for {phot_system}.')
        r = requests.get(url, stream=True)
        with open(tarball, "wb") as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        logger.info(f'Extracting grid from {os.path.basename(url)}.')
        with tarfile.open(tarball) as tar:
            tar.extractall(mist_path)
        os.remove(tarball)



#########################################################################################
# Helper class for reading MIST iso.cmd files
# Slightly modified from 
# https://github.com/jieunchoi/MIST_codes/blob/master/scripts/read_mist_models.py
#########################################################################################
import numpy as np

__all__ = ['IsoReader', 'IsoCmdReader']


class IsoReader(object):

    """

    Reads in MIST isochrone files.

    """

    def __init__(self, filename, verbose=False):

        """

        Args:
            filename: the name of .iso file.

        Usage:
            >> iso = read_mist_models.ISO('MIST_v1.0_feh_p0.00_afe_p0.0_vvcrit0.4.iso')
            >> age_ind = iso.age_index(8.0)
            >> logTeff = iso.isos[age_ind]['log_Teff']
            >> logL = iso.isos[age_ind]['log_L']

        Attributes:
            version     Dictionary containing the MIST and MESA version numbers.
            abun        Dictionary containing Yinit, Zinit, [Fe/H], and [a/Fe] values.
            rot         Rotation in units of surface v/v_crit.
            ages        List of ages.
            num_ages    Number of isochrones.
            hdr_list    List of column headers.
            isos        Data.

        """

        self.filename = filename
        if verbose:
            print('Reading in: ' + self.filename)

        self.version, self.abun, self.rot, self.ages, self.num_ages, self.hdr_list, self.isos = self.read_iso_file()

    def read_iso_file(self):

        """
        Reads in the isochrone file.

        Args:
            filename: the name of .iso file.

        """

        #open file and read it in
        with open(self.filename) as f:
            content = [line.split() for line in f]
        version = {'MIST': content[0][-1], 'MESA': content[1][-1]}
        abun = {content[3][i]:float(content[4][i]) for i in range(1,5)}
        rot = float(content[4][-1])
        num_ages = int(content[6][-1])

        #read one block for each isochrone
        iso_set = []
        ages = []
        counter = 0
        data = content[8:]
        for i_age in range(num_ages):
            #grab info for each isochrone
            num_eeps = int(data[counter][-2])
            num_cols = int(data[counter][-1])
            hdr_list = data[counter+2][1:]
            formats = tuple([np.int32]+[np.float64 for i in range(num_cols-1)])
            iso = np.zeros((num_eeps),{'names':tuple(hdr_list),'formats':tuple(formats)})
            #read through EEPs for each isochrone
            for eep in range(num_eeps):
                iso_chunk = data[3+counter+eep]
                iso[eep]=tuple(iso_chunk)
            iso_set.append(iso)
            ages.append(iso[0][1])
            counter+= 3+num_eeps+2
        return version, abun, rot, ages, num_ages, hdr_list, iso_set

    def age_index(self, age):

        """
        Returns the index for the user-specified age.

        Args:
            age: the age of the isochrone.

        """

        diff_arr = abs(np.array(self.ages) - age)
        age_index = np.where(diff_arr == min(diff_arr))[0][0]

        if ((age > max(self.ages)) | (age < min(self.ages))):
            print('The requested age is outside the range. Try between ' + str(min(self.ages)) + ' and ' + str(max(self.ages)))

        return age_index


            
class IsoCmdReader(object):
    """
    Reads in MIST CMD files.
    """
    
    def __init__(self, filename, verbose=False):
        """
        
        Parameters
        -----------
        filename: the name of .iso.cmd file.
         
        Usage:
            >> isocmd = IsoCmdReader('MIST_v1.0_feh_p0.00_afe_p0.0_vvcrit0.4.iso.cmd')
            >> age_ind = isocmd.age_index(7.0)
            >> B = isocmd.isocmds[age_ind]['Bessell_B']
            >> V = isocmd.isocmds[age_ind]['Bessell_V']
        
        Attributes:
            version         Dictionary containing the MIST and MESA version numbers.
            photo_sys       Photometric system. 
            abun            Dictionary containing Yinit, Zinit, [Fe/H], and [a/Fe] values.
            Av_extinction   Av for CCM89 extinction.
            rot             Rotation in units of surface v/v_crit.
            ages            List of ages.
            num_ages        Number of ages.
            hdr_list        List of column headers.
            isocmds         Data.
        """
        
        self.filename = filename
        if verbose:
            print('Reading in: ' + self.filename)
            
        self.version, self.photo_sys, self.abun, self.Av_extinction, self.rot, self.ages, self.num_ages, self.hdr_list, self.isocmds = self.read_isocmd_file()
    
    def read_isocmd_file(self):

        """

        Reads in the cmd file.
        
        Args:
            filename: the name of .iso.cmd file.
        
        """
        
        #open file and read it in
        with open(self.filename) as f:
            content = [line.split() for line in f]
        version = {'MIST': content[0][-1], 'MESA': content[1][-1]}
        photo_sys = ' '.join(content[2][4:])
        abun = {content[4][i]:float(content[5][i]) for i in range(1,5)}
        rot = float(content[5][-1])
        num_ages = int(content[7][-1])
        Av_extinction = float(content[8][-1])
        
        #read one block for each isochrone
        isocmd_set = []
        ages = []
        counter = 0
        data = content[10:]
        for i_age in range(num_ages):
            #grab info for each isochrone
            num_eeps = int(data[counter][-2])
            num_cols = int(data[counter][-1])
            hdr_list = data[counter+2][1:]
            formats = tuple([np.int32]+[np.float64 for i in range(num_cols-1)])
            isocmd = np.zeros((num_eeps),{'names':tuple(hdr_list),'formats':tuple(formats)})
            #read through EEPs for each isochrone
            for eep in range(num_eeps):
                isocmd_chunk = data[3+counter+eep]
                isocmd[eep]=tuple(isocmd_chunk)
            isocmd_set.append(isocmd)
            ages.append(isocmd[0][1])
            counter+= 3+num_eeps+2
        return version, photo_sys, abun, Av_extinction, rot, ages, num_ages, hdr_list, isocmd_set

    def age_index(self, age):
        
        """

        Returns the index for the user-specified age.
        
        Args:
            age: the age of the isochrone.
        
        """
        
        diff_arr = abs(np.array(self.ages) - age)
        age_index = np.where(diff_arr == min(diff_arr))[0][0]
        
        if ((age > max(self.ages)) | (age < min(self.ages))):
            print('The requested age is outside the range. Try between ' + str(min(self.ages)) + ' and ' + str(max(self.ages)))
            
        return age_index




# Standard library
import os
import pickle
from glob import glob

# Third-party
import numpy as np
from astropy import units as u
from astropy.table import Table

# Project
from . import data_dir


__all__ = ['phot_system_list',
           'FilterSystem',
           'get_filter_names',
           'get_filter_properties',
           'phot_system_lookup',
           'load_zero_point_converter']


# list of photometric systems with pre-calculated filter properties
phot_system_list = [
    'HST_WFC3', 'HST_ACSWF', 'SDSSugriz', 'CFHTugriz', 'DECam', 'HSC',
    'JWST', 'LSST', 'UBVRIplus', 'UKIDSS', 'WFIRST'
]


class FilterSystem(object):
    """
    Class for calculating filter parameters from throughput curves.
    The parameter definitions are taken from
    `Fukugita et al. (1996) AJ 111, 1748.
    <https://ui.adsabs.harvard.edu/abs/1996AJ....111.1748F/abstract>`_

    Parameters
    ----------
    filter_curve_files : list
        List of the file names of the filter curves.
    filter_names : str
        Names of filters. Must be the same length as ``filter_curve_files``.
    **kwargs
        Optional argumnets for `~numpy.loadtxt`, which is used to load the
        filter curve files.
    """

    def __init__(self, filter_curve_files, filter_names, **kwargs):

        self.filter_names = filter_names
        for fn, name in zip(filter_curve_files, filter_names):
            data = np.loadtxt(fn, **kwargs)
            table = Table(data=data, names=['wave', 'trans'])
            setattr(self, name, table)
            self.filter_names.append(name)

    def _get_trans(self, bandpass):
        """
        Get the filter throughput curve for the given bandpass.
        """
        lam = getattr(self, bandpass)['wave']
        trans = getattr(self, bandpass)['trans']
        return lam, trans

    def effective_throughput(self, bandpass):
        """
        Calculate the effective throughput for the given bandpass.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
        theff : float
            Effective effective throughput.
        """
        lam, trans = self._get_trans(bandpass)
        theff = np.trapz(trans, np.log(lam))
        return theff

    def lam_eff(self, bandpass):
        """
        Calculate effective wavelength for the given bandpass.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
            leff : float
                The effective wavelength.
        """
        lam, trans = self._get_trans(bandpass)
        log_leff = np.trapz(np.log(lam) * trans, np.log(lam))
        log_leff /= np.trapz(trans, np.log(lam))
        leff =  np.exp(log_leff) * u.angstrom
        return leff

    def lam_pivot(self, bandpass):
        """
        Calculate pivot wavelength for the given bandpass.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
            lpivot : float
                The pivot wavelength.
        """
        lam, trans = self._get_trans(bandpass)
        lpivot = np.trapz(lam * trans, lam)
        lpivot /= np.trapz(trans, np.log(lam))
        lpivot = np.sqrt(lpivot) * u.angstrom
        return lpivot

    def dlam(self, bandpass):
        """
        Calculate the bandpass width.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
            width : float
                The bandpass width.
        """
        lam, trans = self._get_trans(bandpass)
        norm = np.trapz(trans, lam)
        width = (norm / trans.max()) * u.angstrom
        return width


class ZeroPointConverter(object):
    """
    Help class for converting between AB, ST, and Vega magnitudes & colors.

    Parameters
    ----------
    zpt_table : `~astropy.table.Table`
        MIST zero point table, which can be `downloaded here
        <http://waps.cfa.harvard.edu/MIST/BC_tables/zeropoints.txt>`_.
    """

    def __init__(self, zpt_table):
        for f, system, v_to_st, v_to_ab in zpt_table:
            setattr(self, f, [system, v_to_st, v_to_ab])
        self.zpt_table = zpt_table

    def to_vega(self, bandpass):
        """
        Convert to Vega magnitudes.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        system, v_to_st, v_to_ab = getattr(self, bandpass)
        if system == 'Vega':
            zpt_convert = 0.
        elif system == 'AB':
            zpt_convert = -v_to_ab
        return zpt_convert

    def to_ab(self, bandpass):
        """
        Convert to AB magnitudes.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        system, v_to_st, v_to_ab = getattr(self, bandpass)
        if system == 'AB':
            zpt_convert = 0.
        elif system == 'Vega':
            zpt_convert = v_to_ab
        return zpt_convert

    def to_st(self, bandpass):
        """
        Convert to ST magnitudes.

        Parameters
        ----------
        bandpass : str
            The name of the filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        system, v_to_st, v_to_ab = getattr(self, bandpass)
        if system == 'AB':
            zpt_convert = v_to_st - v_to_ab
        elif system == 'Vega':
            zpt_convert = v_to_st
        return zpt_convert

    def color_to_vega(self, blue, red):
        """
        Convert to Vega colors.

        Parameters
        ----------
        blue : str
            The name of the blue filter.
        red : str
            The name of the red filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        blue_convert = self.to_vega(blue)
        red_convert = self.to_vega(red)
        return blue_convert - red_convert

    def color_to_ab(self, blue, red):
        """
        Convert to AB colors.

        Parameters
        ----------
        blue : str
            The name of the blue filter.
        red : str
            The name of the red filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        blue_convert = self.to_ab(blue)
        red_convert = self.to_ab(red)
        return blue_convert - red_convert

    def color_to_st(self, blue, red):
        """
        Convert to ST colors.

        Parameters
        ----------
        blue : str
            The name of the blue filter.
        red : str
            The name of the red filter.

        Returns
        -------
        zpt_convert : float
            Zero point conversion magnitude.
        """
        blue_convert = self.to_st(blue)
        red_convert = self.to_st(red)
        return blue_convert - red_convert


def get_filter_names(phot_system=None):
    """
    Get MIST photometric systems and filter names.

    Parameters
    ----------
    phot_system : str or None
        The desired photometric system.

    Returns
    -------
    mist_filter_names : list or dict
        If ``phot_system`` given, then a list of filter names is returned. If
        ``phot_system`` is ``None``, then a dict of all photometric systems
        and filters is returned.
    """
    fn = os.path.join(data_dir, 'mist_filter_names.pkl')
    pickle_in = open(fn, 'rb')
    mist_filter_names = pickle.load(pickle_in)
    pickle_in.close()
    if phot_system is not None:
        if type(phot_system) == str:
            phot_system = [phot_system]
        names = []
        for p in phot_system:
            names.extend(mist_filter_names[p])
        mist_filter_names = names
    return mist_filter_names


def get_filter_properties():
    """

    """
    pkl_fn = os.path.join(data_dir, 'filter_properties.pkl')
    with open(pkl_fn, 'rb') as pkl_file:
        data = pickle.load(pkl_file)
        props = Table(data)
        props['dlam'] *= u.angstrom
        props['lam_eff'] *= u.angstrom
    return props


def phot_system_lookup(filter_name=None):
    """
    Lookup the photometric system name associated with a given filter name.

    Parameters
    ----------
    filter_name : str, optional
        Filter name to lookup.

    Returns
    -------
    lookup : dict or str
        If ``filter_name`` is ``None``, a dictionary with the filter names as
        keywords and the photometric systems as values. If ``filter_name`` is
        not ``None``, its photometric system is returned.
    """
    fn = os.path.join(data_dir, 'phot_system_lookup.pkl')
    pickle_in = open(fn, 'rb')
    lookup = pickle.load(pickle_in)
    pickle_in.close()
    if filter_name is not None:
        lookup = lookup[filter_name]
    return lookup


def load_zero_point_converter():
    """
    Create and return a `~artpop.filters.ZeroPointConverter` object.
    """
    from astropy.io import ascii
    fn = os.path.join(data_dir, 'zeropoints.txt')
    table= ascii.read(fn)
    return ZeroPointConverter(table)



# Third-party
import numpy as np
from astropy import units as u
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.convolution import Gaussian2DKernel, Moffat2DKernel

# Project
from ..util import check_units, check_odd


__all__ = ['gaussian_psf', 'moffat_psf']


def _check_shape(shape):
    """Make sure shape is a tuple."""
    if type(shape) == int:
        shape = (shape, shape)
    check_odd(shape, 'psf shape')
    return shape


def gaussian_psf(fwhm, pixel_scale=0.2, shape=41):
    """
    Gaussian point-spread function.

    Parameters
    ----------
    fwhm : float or `~astropy.units.Quantity`
        Full width at half maximum of the psf. If a float is given, the units
        will be assumed to be `~astropy.units.arcsec`. The units can be angular 
        or in pixels.
    pixel_scale : float or `~astropy.units.Quantity`, optional
        The pixel scale of the psf image. If a float is given, the units are 
        assumed to be `~astropy.units.arcsec` per `~astropy.units.pixel` 
        (why would you want anything different?).
    shape : int or list-like, optional
        Shape of the psf image. Must be odd. If an int is given, the x and y 
        dimensions will be set to this value: (shape, shape).

    Returns
    -------
    psf : `~numpy.ndarray`
        The PSF image normalized such that its sum is equal to one. 
    """

    fwhm = check_units(fwhm, 'arcsec')
    pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)
    y_size, x_size = _check_shape(shape)

    width = fwhm.to('pixel', u.pixel_scale(pixel_scale)).value
    width *= gaussian_fwhm_to_sigma
    model = Gaussian2DKernel(x_stddev=width,
                             y_stddev=width,
                             x_size=x_size,
                             y_size=y_size)
    model.normalize()
    psf = model.array
    return psf


def moffat_psf(fwhm, pixel_scale=0.2, shape=41, alpha=4.765):
    """
    Moffat point-spread function.

    Parameters
    ----------
    fwhm : float or `~astropy.units.Quantity`
        Full width at half maximum of the psf. If a float is given, the units
        will be assumed to be arcsec. The units can be angular or in pixels.
    pixel_scale : float or `~astropy.units.Quantity`, optional
        The pixel scale of the psf image. If a float is given, the units are 
        assumed to be arcsec / pixel (why would you want anything different?).
    shape : int or list-like, optional
        Shape of the psf image. Must be odd. If an int is given, the x and y 
        dimensions will be set to this value: (shape, shape).
    alpha : float, optional
        Power index of the Moffat model.

    Returns
    -------
    psf : `~numpy.ndarray`
        The PSF image normalized such that its sum is equal to one. 

    Notes
    -----
    The default value `alpha = 4.765` is a fit to the prediction from 
    atmospheric turbulence theory (`Trujillo et al. 2001 
    <https://ui.adsabs.harvard.edu/abs/2001MNRAS.328..977T/abstract>`_).
    """
    fwhm = check_units(fwhm, 'arcsec')
    pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)
    y_size, x_size = _check_shape(shape)

    width = fwhm.to('pixel', u.pixel_scale(pixel_scale)).value
    gamma = width / (2 * np.sqrt(2**(1 / alpha) - 1))
    model = Moffat2DKernel(gamma=gamma,
                           alpha=alpha,
                           x_size=x_size,
                           y_size=y_size)
    model.normalize()
    psf = model.array
    return psf



__version__ = "0.1.0"

try:
    __ARTPOP_SETUP__
except NameError:
    __ARTPOP_SETUP__ = False

if not __ARTPOP_SETUP__:
    import os
    package_dir = os.path.dirname(__file__)
    data_dir = os.path.join(package_dir, 'data')
    MIST_PATH = os.getenv('MIST_PATH')
    if MIST_PATH is None:
        MIST_PATH = os.path.join(os.path.expanduser('~'), '.artpop')
        if not os.path.exists(MIST_PATH):
            os.mkdir(MIST_PATH)
        MIST_PATH = os.path.join(MIST_PATH, 'mist')
        if not os.path.exists(MIST_PATH):
            print('\033[33mWARNING:\033[0m MIST_PATH environment variable does'
                  ' not exist. When you first use a MIST grid, it will be '
                 f'saved in\n{MIST_PATH}. To change this '
                  'location, create a MIST_PATH environment variable.')
            os.mkdir(MIST_PATH)
    from .filters import *
    from .stars import *
    from .image import *
    from .space import *
    from .source import *
    from .visualization import *



# Third-party
import numpy as np
from astropy.modeling import Fittable2DModel, Parameter


__all__ = ['Plummer2D', 'Constant2D']


class Plummer2D(Fittable2DModel):
    """
    Two-dimensional Plummer surface brightness profile.

    Parameters
    ----------
    amplitude : float
        Central surface brightness.
    scale_radius : float
        Characteristic scale radius of the mass distribution.
    x_0 : float, optional
        x position of the center.
    y_0 : float, optional
        y position of the center.
    """

    amplitude = Parameter(default=1)
    scale_radius = Parameter(default=1)
    x_0 = Parameter(default=0)
    y_0 = Parameter(default=0)

    @classmethod
    def evaluate(cls, x, y, amplitude, scale_radius, x_0, y_0):
        """Two-dimensional Plummer profile evaluation function."""
        r = np.sqrt((x - x_0)**2 + (y - y_0)**2)
        return amplitude / (1 + (r / scale_radius)**2)**2


class Constant2D(Fittable2DModel):
    """
    The simplest model ever.

    Parameters
    ----------
    amplitude : float
        Constant surface brightness.
    """

    amplitude = Parameter(default=1)

    @classmethod
    def evaluate(cls, x, y, amplitude):
        """Constant surface brightness evaluation function."""
        if hasattr(x, 'shape') and hasattr(y, 'shape'):
            assert x.shape == y.shape, 'Shapes of x and y must match'
            z = np.ones(x.shape)
        elif hasattr(x, 'shape'):
            z = np.ones(x.shape)
        elif hasattr(y, 'shape'):
            z = np.ones(y.shape)
        else:
            z = 1.0
        return amplitude * z



# Standard library
import os
import sys
import abc
import pickle
from copy import deepcopy

# Third-party
import numpy as np
from astropy.io import fits
from astropy import constants
from astropy import units as u
from astropy.convolution import convolve_fft
from fast_histogram import histogram2d

# Project
from .. import data_dir
from ..util import check_units, check_random_state
from ..filters import FilterSystem, get_filter_names, get_filter_properties
from ..source import Source
from ..log import logger


__all__ = ['IdealObservation', 'ArtObservation', 'IdealImager', 'ArtImager']


class Observation(metaclass=abc.ABCMeta):

    def to_pickle(self, file_name):
        """Pickle observation object."""
        pkl_file = open(file_name, 'wb')
        pickle.dump(self, pkl_file)
        pkl_file.close()

    @staticmethod
    def from_pickle(file_name):
        """Load pickle of observation object."""
        pkl_file = open(file_name, 'rb')
        data = pickle.load(pkl_file)
        pkl_file.close()
        return data

    def to_fits(self, file_name, overwrite=True, image_type='image', **kwargs):
        """
        Write image to fits file.

        Parameters
        ----------
        file_name : str
            Name of fits file.
        overwrite : str, optional
            If True (default), overwrite the image if it exists.
        image_type : str, optional
            Attribute name of the image to be written.
        """
        fits.writeto(file_name,
                     getattr(self, image_type),
                     overwrite=overwrite,
                     **kwargs)

    def copy(self):
        """Create deep copy of observation object."""
        return deepcopy(self)


class IdealObservation(Observation):
    """
    Return object for the ideal imager.

    Parameters
    ----------
    image : `~numpy.ndarray`
        The ideal mock image.
    bandpass : str
        Filter of the observation.
    zpt : float
        The magnitude zero point.
    """
    def __init__(self, image, zpt, bandpass):
        self.image = image
        self.zpt = zpt
        self.bandpass = bandpass


class ArtObservation(Observation):
    """
    Return object for the artificial imager.

    Parameters
    ----------
    raw_counts : `~numpy.ndarray`
        Raw count image, including Poisson noise.
    src_counts : `~numpy.ndarray`
        Source count image before Poission noise is added.
    sky_counts : float
        Counts per pixel from the sky.
    var_image : `~numpy.ndarray`
        Variance image.
    calibration : float
        Calibration factor that converts counts to calibrated flux units.
    bandpass : str
        Filter of the observation.
    zpt : float
        The magnitude zero point.
    exptime : `~astropy.units.Quantity`
        The exposure time of the mock observation.
    """
    def __init__(self, raw_counts, src_counts, sky_counts, image,
                 var_image, calibration, zpt, bandpass, exptime, mag_error):
        self.raw_counts = raw_counts
        self.src_counts = src_counts
        self.sky_counts = sky_counts
        self.image = image
        self.var_image = var_image
        self.calibration = calibration
        self.zpt = zpt
        self.bandpass = bandpass
        self.exptime = exptime
        self.mag_error = mag_error


mAB_0 = 48.6
def fnu_from_AB_mag(mag):
    """
    Convert AB magnitude into flux density fnu in cgs units.
    """
    fnu = 10.**((mag + mAB_0)/(-2.5))
    return fnu*u.erg/u.s/u.Hz/u.cm**2


class Imager(metaclass=abc.ABCMeta):
    """Base class for Imager objects."""

    def _check_source(self, source):
        """Verify that input object is a Source object."""
        if Source not in source.__class__.__mro__:
            raise Exception(f'{type(src)} is not a valid Source object.')

    def _check_bandpass(self, bandpass):
        """Verify bandpass is the loaded filter list."""
        if bandpass not in self.filters:
            raise Exception(f'{bandpass} filter properties were not provided.')

    def inject_stars(self, x, y, signal, xy_dim, mask=None):
        """
        Inject sources into image.

        Parameters
        ----------
        x : `~numpy.ndarray`
            Source x coordinates.
        y : `~numpy.ndarray`
            Source y coordinates.
        signal : `~numpy.ndarray`
            The signal to inject into image (e.g., flux, counts, etc...).
        xy_dim : int or list-like
            The dimensions of mock image in xy units. If `int` is given, it is
            assumed to be both the x and y dimensions: (xy_dim, xy_dim).
        mask : `~numpy.ndarray`, optional
            Boolean mask that is set to True for stars you want to inject.
            If None, all stars will be injected.

        Returns
        -------
        image : `~numpy.ndarray`
            The mock image *before* psf convolution.
        """
        bins = tuple(np.asarray(xy_dim).astype(int))
        hist_range = [[0, bins[0] - 1], [0, bins[1]- 1]]
        if mask is not None:
            _x = x[mask]
            _y = y[mask]
            _s = signal[mask]
        else:
            _x, _y, _s = x, y, signal
        image = histogram2d(_x, _y, bins=bins, weights=_s, range=hist_range).T
        return image

    def apply_seeing(self, image, psf=None, boundary='wrap'):
        """
        Convolve mock image with psf.

        Parameters
        ----------
        image : `~numpy.ndarray`
            The mock image.
        psf : `~numpy.ndarray` or None, optional
            The point-spread function. If None, will return the input image.
        boundary : {'fill', 'wrap'}, optional
            A flag indicating how to handle boundaries:

                * 'fill': set values outside the array boundary to fill_value
                  (default)
                * 'wrap': periodic boundary

            The `None` and 'extend' parameters are not supported for FFT-based
            convolution.

        Returns
        -------
        image : `~numpy.ndarray`
            The psf-convolved image. If `psf` is not `None`, the original image
            is returned.
        """
        if psf is not None:
            image = convolve_fft(image , psf,
                                 boundary=boundary,
                                 normalize_kernel=True)
        return image

    def inject_smooth_model(self, bimage, source, bandpass, zpt):
        """Inject smooth model if it exists."""
        return NotImplementedError()

    def observe(self, bandpass, psf, **kwargs):
        """Mock observe source."""
        return NotImplementedError()


class IdealImager(Imager):
    """Ideal imager for making noise-free images."""

    def inject_smooth_model(self, image, source, bandpass, zpt):
        """
        Inject smooth component of the star system into image if it exists.

        Parameters
        ----------
        image : `~numpy.ndarray`
            The artificial image.
        source : `~artpop.source.Source`
            The artificial source.
        bandpass : str
            Observational bandpass of the mock observation.
        zpt : float
            Photometric zero point.

        Returns
        -------
        image : `~numpy.ndarray`
            The artificial image with the smooth model injected into it. If no
            smooth model exists, the original image will be returned.
        """
        if hasattr(source, 'smooth_model'):
            if source.smooth_model is None:
                image = image
            else:
                mag = source.sp.mag_integrated_component(bandpass)
                amp, amp_image, name = source.mag_to_image_amplitude(mag, zpt)
                setattr(source.smooth_model, name, amp_image)
                yy, xx = np.mgrid[:image.shape[0], :image.shape[1]]
                image = image + source.smooth_model(xx, yy)
        else:
            image = image
        return image

    def observe(self, source, bandpass, psf=None, zpt=27, mask=None, **kwargs):
        """
        Make ideal observation.

        Parameters
        ----------
        source : `~artpop.source.Source`
            Source to mock observer.
        bandpass : str
            Filter of observation. Must be a filter in the given
            photometric system(s).
        psf : `~numpy.ndarray` or None, optional
            The point-spread function. If None, will not psf-convolve image.
        zpt : float, optional
            The magnitude zero point of the mock image.
        mask : `~numpy.ndarray`, optional
            Boolean mask that is set to True for stars you want to inject.
            If None, all stars will be injected.

        Returns
        -------
        observation : `~artpop.image.IdealObservation`
            Observation object for the ideal imager.
        """
        self._check_source(source)
        flux = 10**(0.4 * (zpt - source.mags[bandpass]))
        image = self.inject_stars(source.x, source.y, flux,
                                  source.xy_dim, mask)
        image = self.inject_smooth_model(image, source, bandpass, zpt)
        image = self.apply_seeing(image, psf, **kwargs)
        observation = IdealObservation(image=image, bandpass=bandpass, zpt=zpt)
        return observation


class ArtImager(Imager):
    """
    Imager for making fully artificial images.

    .. note::
        If you use a phot_system and its pre-calculated filter properties, the
        conversion from magnitude to counts assumes AB magnitudes.

    Parameters
    ----------
    phot_system : str or list-like, optional
        Name of the photometric system(s). Use this option if you are using
        magnitudes from one of the systems with pre-calculated filter
        properties included with ArtPop. See `~artpop.phot_system_list`.
    diameter : float or `~astropy.units.Quantity`, optional
        Diameter of the telescope aperture. If a float is given, the units
        will be assumed to be meters.
    read_noise : float, optional
        RMS of Gaussian read noise. Set to zero by default.
    efficiency : float, optional
        Efficiency factor (e.g., to account for CCD efficiency, atmospheric
        transmission, etc). It is set to one by default.
    dlam : dict of floats or dict of ~astropy.units.Quantity`, optional
        Dictionary containing the filter widths. Used for converting magnitudes
        into counts. Filter names should be the keys and the widths the values.
        If the wavelengths are given as floats, angstroms will be assumed.
    lam_eff : dict of floats or dict of ~astropy.units.Quantity`, optional
        Dictionary containing the filter effective wavelengths. Used for
        converting magnitudes into counts. Filter names should be the keys
        and the widths should be the values. If the wavelengths are given as
        floats, angstroms will be assumed.
    filter_system : `~artpop.FilterSystem`, optional
        Filter system object with filter curves for the filters you use to
        create artificial images. This object calculates `dlam` and `lam_eff`
        for converting magnitudes into counts.
    zpt_inst : dictionary, optional
        Instrumental zero points. Format should be a dictionary with the filter
        names as keys and the zero points as values. If not None, this alone
        sets the magnitude associated with 1 count per second. This parameter
        takes precedence; parameters such as the diameter will be ignored.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    """

    def __init__(self, phot_system=None, diameter=10, read_noise=0.0,
                 efficiency=1.0, dlam=None, lam_eff=None, filter_system=None,
                 zpt_inst=None, random_state=None):

        self.efficiency = efficiency
        self.read_noise = read_noise
        self.diameter = check_units(diameter, 'm')
        self.rng = check_random_state(random_state)

        self.dlam = None
        self.lam_eff = None
        self.filters = None
        self.zpt_inst = None
        if phot_system is not None:
            self.dlam = {}
            self.lam_eff = {}
            self.phot_system = phot_system
            self.filters = get_filter_names(phot_system)
            props = get_filter_properties()
            for filt in self.filters:
                select = props['bandpass'] == filt
                self.dlam[filt] = props[select]['dlam'][0]
                self.lam_eff[filt] = props[select]['lam_eff'][0]
        elif zpt_inst is not None:
            self.filters = list(zpt_inst.keys())
            self.zpt_inst = zpt_inst
        elif dlam is not None:
            assert lam_eff is not None, 'must give both dlam and lam_eff'
            self.filters = list(dlam.keys())
            self.dlam = dlam
            self.lam_eff = lam_eff
        elif filter_system is not None:
            assert type(filter_system) == FilterSystem
            self.filter_system = filter_system
            self.filters = filter_system.filter_names
            for filt in self.filters:
                self.dlam[filt] = filter_system.dlam(filt).value
                self.lam_eff[filt] = filter_system.lam_eff(filt).value
        else:
            msg = 'must give phot_system or (dlam, lam_eff) or filter_system.'
            raise Exception('In order to convert mags to counts, you ' + msg)

        if self.zpt_inst is None:
            for filt in self.filters:
                self.dlam[filt] = check_units(self.dlam[filt], 'angstrom')
                self.lam_eff[filt] = check_units(
                    self.lam_eff[filt], 'angstrom')

    @property
    def area(self):
        """Area of aperture."""
        r = self.diameter / 2
        return np.pi * r**2

    def mag_to_counts(self, mags, bandpass, exptime):
        """
        Convert magnitudes to counts.

        Parameters
        ----------
        mags : `~numpy.ndarray` or `~astropy.table.Column`
            AB magnitudes to be converted into counts.
        bandpass : str
            Filter of observation. Must be a filter in the given
            photometric system(s).
        exptime : float or `~astropy.units.Quantity`
            Exposure time. If float is given, the units are assumed to
            be `~astropy.units.second`.

        Returns
        -------
        counts : `~numpy.ndarray`
            The counts associated with each magnitude.
        """
        exptime = check_units(exptime, 's')
        if self.zpt_inst is None:
            fnu = fnu_from_AB_mag(mags)
            dlam = self.dlam[bandpass]
            lam_eff = self.lam_eff[bandpass]
            photon_flux = (dlam / lam_eff) * fnu / constants.h.to('erg s')
            counts = photon_flux * self.area.to('cm2') * exptime.to('s')
            counts = counts.decompose()
            assert counts.unit == u.dimensionless_unscaled
            counts *= self.efficiency
            counts = counts.value
        else:
            counts = 10**(0.4 * (self.zpt_inst[bandpass] - mags))
            counts *= exptime.to('s').value
        return counts

    def sb_to_counts_per_pixel(self, sb, bandpass, exptime, pixel_scale):
        """
        Convert a constant surface brightness into counts per pixel.

        Parameters
        ----------
        sb : float
            Surface brightness in units of `~astropy.units.mag` per square
            `~astropy.units.arcsec`.
        bandpass : str
            Filter of observation. Must be a filter in the given
            photometric system(s).
        exptime : float or `~astropy.units.Quantity`
            Exposure time. If float is given, the units are assumed to
            be `~astropy.units.second`.
        pixel_scale : `~astropy.units.Quantity`
            Pixel scale.

        Returns
        -------
        counts_per_pixel : float
            Counts per pixel associated with the given surface brightness.
        """
        exptime = check_units(exptime, 's')
        if self.zpt_inst is None:
            fnu_per_square_arcsec = fnu_from_AB_mag(sb) / u.arcsec**2
            dlam = self.dlam[bandpass]
            lam_eff = self.lam_eff[bandpass]
            pixel_scale = pixel_scale.to('arcsec / pixel')
            E_lam = (constants.h * constants.c / lam_eff).decompose().to('erg')
            flam_per_square_arcsec = fnu_per_square_arcsec *\
                constants.c.to('angstrom/s') / lam_eff**2
            flam_per_pixel = flam_per_square_arcsec * pixel_scale**2
            photon_flux_per_sq_pixel = (flam_per_pixel * dlam / E_lam).\
                decompose().to('1/(cm2*pix2*s)')
            counts_per_pixel = photon_flux_per_sq_pixel * exptime.to('s')
            counts_per_pixel *= self.area.to('cm2') * u.pixel**2
            assert counts_per_pixel.unit == u.dimensionless_unscaled
            counts_per_pixel *= self.efficiency
            counts_per_pixel = counts_per_pixel.value
        else:
            counts_per_pixel = 10**(0.4 * (self.zpt_inst[bandpass] - sb))
            counts_per_pixel *= exptime.to('s').value
        return counts_per_pixel

    def calibration(self, bandpass, exptime, zpt=27.0):
        """
        Calculate the calibration factor, which converts counts into
        calibrated flux units.

        Parameters
        ----------
        bandpass : str
            Filter of observation. Must be a filter in the given
            photometric system(s).
        exptime : float or `~astropy.units.Quantity`
            Exposure time. If float is given, the units are assumed to
            be `~astropy.units.second`.
        zpt : float, optional
            The magnitude zero point of the mock image.

        Returns
        -------
        cali_factor : float
            The calibration factor.
        """
        if self.zpt_inst is None:
            dlam_over_lam = self.dlam[bandpass] / self.lam_eff[bandpass]
            lam_factor = (dlam_over_lam / constants.h.cgs).value
            cali_factor = 10**(0.4 * zpt) * 10**(0.4 * mAB_0) / lam_factor
            cali_factor /= exptime.to('s').value
            cali_factor /= self.area.to('cm2').value
            cali_factor /= self.efficiency
        else:
            cali_factor = 10**(0.4 * (zpt - self.zpt_inst[bandpass]))
            cali_factor /= exptime.to('s').value
        return cali_factor

    def inject_smooth_model(self, image, source, bandpass, exptime, zpt):
        """
        Inject smooth component of the star system into image if it exists.

        Parameters
        ----------
        image : `~numpy.ndarray`
            The artificial image.
        source : `~artpop.source.Source`
            The artificial source.
        bandpass : str
            Observational bandpass of the mock observation.
        exptime : float or `~astropy.units.Quantity`
            Exposure time. If `float` is given, the units are assumed to
            be `~astropy.units.second`.
        zpt : float
            Photometric zero point.

        Returns
        -------
        image : `~numpy.ndarray`
            The artificial image with the smooth model injected into it. If no
            smooth model exists, the original image will be returned.
        """
        if hasattr(source, 'smooth_model'):
            if source.smooth_model is None:
                image = image
            else:
                mag = source.sp.mag_integrated_component(bandpass)
                amp, _, name = source.mag_to_image_amplitude(mag, zpt)
                amp_counts = self.sb_to_counts_per_pixel(
                    amp, bandpass, exptime, source.pixel_scale)
                setattr(source.smooth_model, name, amp_counts)
                yy, xx = np.mgrid[:image.shape[0], :image.shape[1]]
                image = image + source.smooth_model(xx, yy)
        else:
            image = image
        return image

    def observe(self, source, bandpass, exptime, sky_sb=None, psf=None,
                zpt=27.0, mask=None, **kwargs):
        """
        Make artificial observation.

        Parameters
        ----------
        source : `~artpop.source.Source`
            Artificial source to be observed.
        bandpass : str
            Filter of observation. Must be a filter in the given
            photometric system(s).
        exptime : float or `~astropy.units.Quantity`
            Exposure time. If `float` is given, the units are assumed to
            be `~astropy.units.second`.
        sky_sb : float or None, optional
            Constant surface brightness of the sky. If None is given, then no
            sky noise will be added.
        psf : `~numpy.ndarray` or None, optional
            The point-spread function. If None, will not psf-convolve image.
        zpt : float, optional
            The magnitude zero point of the mock image.
        mask : `~numpy.ndarray`, optional
            Boolean mask that is set to True for stars you want to inject.
            If None, all stars will be injected.

        Returns
        -------
        observation : `~artpop.image.ArtObservation`
            Observation object for the artificial imager.
        """
        self._check_bandpass(bandpass)
        self._check_source(source)
        exptime = check_units(exptime, 's')
        counts = self.mag_to_counts(source.mags[bandpass], bandpass, exptime)
        src_counts = self.inject_stars(source.x, source.y,
                                       counts, source.xy_dim, mask)
        src_counts = self.inject_smooth_model(src_counts, source,
                                              bandpass, exptime, zpt)
        src_counts = self.apply_seeing(src_counts, psf, **kwargs)
        if sky_sb is not None:
            sky_counts = self.sb_to_counts_per_pixel(
                sky_sb, bandpass, exptime, source.pixel_scale)
            n_s = np.sqrt(self.read_noise**2 + sky_counts + counts) / counts
            mag_error = 2.5 * np.log10(1 + n_s)
        else:
            src_counts[src_counts < 0] = 0
            sky_counts = 0
            mag_error = None
        # HACK: Use Normal distribution for Windows. Poisson sampling
        # breaks in Windows due to long ints being 32 bit.
        if sys.platform =='win32':
            _c = src_counts + sky_counts
            raw_counts = self.rng.normal(loc=_c, scale=np.sqrt(_c))
        else:
            raw_counts = self.rng.poisson(src_counts + sky_counts)
        if self.read_noise > 0.0:
            rn = self.rng.normal(scale=self.read_noise, size=src_counts.shape)
            raw_counts = raw_counts + rn
        cali = self.calibration(bandpass, exptime, zpt)
        image_cali = (raw_counts - sky_counts) * cali
        var = raw_counts + self.read_noise**2
        observation = ArtObservation(
            raw_counts=raw_counts,
            src_counts=src_counts,
            sky_counts=sky_counts,
            image=image_cali,
            var_image=var,
            calibration=cali,
            zpt=zpt,
            bandpass=bandpass,
            exptime=exptime,
            mag_error=mag_error
        )
        return observation



# Standard library
import sys
import logging

# Third-party
from astropy.logger import AstropyLogger


class ArtPopLogger(AstropyLogger):

    def _set_defaults(self, level='INFO'):
        """
        Reset logger to its initial state
        """
        # Remove all previous handlers
        for handler in self.handlers[:]:
            self.removeHandler(handler)

        # Set levels
        self.setLevel(level)

        # Set up the stdout handler
        sh = StreamHandler()
        self.addHandler(sh)


class StreamHandler(logging.StreamHandler):
    """
    A specialized StreamHandler that logs INFO and DEBUG messages to
    stdout, and all other messages to stderr.  Also provides coloring
    of the output, if enabled in the parent logger.
    """

    def emit(self, record):
        '''
        The formatter for stderr
        '''
        # Import utils.console only if necessary and at the latest because
        # the import takes a significant time [#4649]
        from astropy.utils.console import color_print

        if record.levelno <= logging.INFO:
            stream = sys.stdout
        else:
            stream = sys.stderr

        if record.levelno < logging.INFO:
            color_print(record.levelname, 'magenta', end='', file=stream)
        elif record.levelno < logging.WARN:
            color_print(record.levelname, 'green', end='', file=stream)
        elif record.levelno < logging.ERROR:
            color_print(record.levelname, 'brown', end='', file=stream)
        else:
            color_print(record.levelname, 'red', end='', file=stream)
        record.message = "{0}".format(record.msg)
        print(": " + record.message, file=stream)


logging.setLoggerClass(ArtPopLogger)
logger = logging.getLogger('ArtPop Logger')
logger._set_defaults()



# Standard library
import os

# Third-party
import numpy as np
from numpy.lib.recfunctions import append_fields
from scipy.integrate import quad
from scipy.interpolate import interp1d
from astropy.table import Table
from astropy import units as u

# Project
from ._read_mist_models import IsoCmdReader, IsoReader
from .imf import imf_dict, IMFIntegrator
from .. import MIST_PATH
from ..log import logger
from ..filters import phot_system_list, get_filter_names
from ..filters import load_zero_point_converter
from ..util import check_units, fetch_mist_grid_if_needed
phot_str_helper = {p.lower():p for p in phot_system_list}


__all__ = ['fetch_mist_iso_cmd', 'Isochrone', 'MISTIsochrone']


class Isochrone(object):
    """
    Class for storing generic isochrones. It also has methods for calculating
    IMF-weighted properties of a simple stellar population with the given
    age and metallicity.

    Parameters
    ----------
    mini : `~numpy.ndarray`
        Initial stellar masses. The masses must be monotonically increasing.
    mact : `~numpy.ndarray`
        Actual stellar masses after mass loss.
    mag_table : `~astropy.table.Table`, dict, or structured `~numpy.ndarray`
        The stellar magnitudes.
    eep : `~numpy.ndarray`, optional
        The Primary Equivalent Evolutionary Points. Needed to identify phases
        of stellar evolution.
    log_L : `~numpy.ndarray`, optional
        Stellar luminosities.
    log_Teff : `~numpy.ndarray`, optional
        Stellar effective temperatures.
    """

    def __init__(self, mini, mact, mags, eep=None, log_L=None,
                 log_Teff=None):
        self.mini = np.asarray(mini)
        self.mact = np.asarray(mact)
        if (np.diff(self.mini) < 0).sum() > 0:
            raise Exception('Initial masses must be monotonically increasing.')
        self.eep = None if eep is None else np.asarray(eep)
        self.log_L = None if log_L is None else np.asarray(log_L)
        self.log_Teff = None if log_Teff is None else np.asarray(log_Teff)
        if type(mags) == dict or type(mags) == np.ndarray:
            self.mag_table = Table(mags)
        elif type(mags) == Table:
            self.mag_table = mags
        else:
            raise Exception(f'{type(mags)} is not a valid type for mags')

    @property
    def m_min(self):
        """The minimum mass of the isochrone."""
        return self.mini.min()

    @property
    def m_max(self):
        """The maximum mass of the isochrone."""
        return self.mini.max()

    @property
    def filters(self):
        """List of filters in the given photometric system(s)."""
        return self.mag_table.colnames

    @staticmethod
    def from_parsec(file_name, log_age=None, zini=None, num_filters=None):
        """
        Read isochrone generated from the `PARSEC website
        <http://stev.oapd.inaf.it/cgi-bin/cmd>`_. If more than one age and/or
        metallicity is included in the file, you must provide the log_age
        and/or zini parameters.


        Parameters
        ----------
        file_name : str
            Isochrone file name.
        log_age : float, optional
            Log of age in years. You must provided this parameter if there is
            more than one age in the isochrone file. Note this function does
            not interpolate ages, so it must be included in the file.
        zini : float, optional
            Initial metal fraction. You must provided this parameter if there
            is more than one metallicity in the isochrone file. Note this
            function does not interpolate metallicity, so it must be included
            in the file.
        num_filters : int, optional
            Number of filters included in the isochrone file. If None, will
            assume the last non-filter parameter is `mbolmag`.

        Returns
        -------
        iso : `artpop.stars.Isochrone`
            PARSEC Isochrone object.
        """
        if os.path.isfile(file_name):
            file = open(file_name, 'r')
            lines = file.readlines()
            file.close()
            for l in lines:
                if 'Zini' in l.split()[1]:
                    names = l.split()[1:]
                    break
            data = np.loadtxt(file_name)
            parsec = Table(data, names=names)
        else:
            raise Exception(f'{file_name} does not exist.')
        isochrone_full = parsec.copy()
        if log_age is not None:
            age_cut = np.abs(parsec['logAge'] - log_age) < 1e-5
            if age_cut.sum() < 1:
                raise Exception(f'log_age = {log_age} not found.')
            parsec = parsec[age_cut]
        if zini is not None:
            zini_cut = np.abs(parsec['Zini'] - zini) < 1e-8
            if zini_cut.sum() < 1:
                raise Exception(f'Zini = {zini} not found.')
            parsec = parsec[zini_cut]
        if num_filters is None:
            filt_idx = np.argwhere(np.array(names) == 'mbolmag')[0][0] + 1
        else:
            filt_idx = len(names) - num_filters
        iso = Isochrone(mini=parsec['Mini'],
                        mact=parsec['Mass'],
                        mags=parsec[names[filt_idx:]],
                        log_L=parsec['logL'],
                        log_Teff=parsec['logTe'])
        iso.isochrone_full = isochrone_full
        return iso

    def interpolate(self, y_name, x_interp, x_name='mini',
                    slice_interp=np.s_[:], **kwargs):
        """
        Interpolate isochrone.

        Parameters
        ----------
        y_name : str
            Parameter name of interpolated variable.
        x_interp : `~numpy.ndarray`
            New x values to interpolate over.
        x_name : str
            Parameter name of x values. Usually want this to be initial mass.

        Returns
        -------
        y_interp : `~numpy.ndarray`
            Interpolated y values.
        """
        x = self.mag_table[x_name] if x_name in self.filters\
                                   else getattr(self, x_name)
        y = self.mag_table[y_name] if y_name in self.filters\
                                   else getattr(self, y_name)
        x = x[slice_interp]
        y = y[slice_interp]
        y_interp = interp1d(x, y, **kwargs)(x_interp)
        return y_interp

    def mag_to_mass(self, mag, bandpass):
        """
        Interpolate isochrone to calculate initial stellar masses that have
        the given magnitude. Since more than one mass can have the same
        magnitude (from stars being in different phases of stellar evolution),
        we step through each isochrone and check if the given magnitude falls
        between two bins, in which case we interpolate the associated mass.

        Parameters
        ----------
        mag : float
            Magnitude to be converted in to stellar mass(es).
        bandpass : str
            Photometric bandpass of the `mag`.

        Returns
        -------
        mass_interp : `~numpy.ndarray`
            Interpolated mass(es) associated with `mag`.
        """
        y = self.mini
        x = self.mag_table[bandpass]

        if mag < x.min() or mag > x.max():
            raise Exception(f'mag = {mag} is outside the isochrone range.')

        y_interp = []
        # Loop over isochrone and check if desired magnitude is in between two
        # bins. If it is, interpolate the mass. Looping like this is necessary
        # because different masses can have the same magnitude due to
        # being at different phases of stellar evolution.
        for i in range(len(x)):
            if x[i] == mag:
                y_interp.append(y[i])
                continue
            if i < len(x) - 1:
                _x = [x[i], x[i + 1]]
                _y = [y[i], y[i + 1]]
                dx = x[i + 1] - x[i]
                if dx > 0:
                    if (mag <= x[i + 1]) and (mag > x[i]):
                        y_interp.append(interp1d(_x, _y)([mag])[0])
                else:
                    if (mag >= x[i + 1]) and (mag < x[i]):
                        y_interp.append(interp1d(_x, _y)([mag])[0])
        mass_interp = np.array(y_interp)

        return mass_interp

    def calculate_mag_limit(self, imf, bandpass, frac_mass_sampled=None,
                            frac_num_sampled=None, distance=10*u.pc):
        """
        Calculate the limiting faint magnitude to sample a given fraction
        of mass or number of a stellar population. This is used for when you
        only what to sample stars that are brighter than a magnitude limit.
        You must specify either `frac_mass_sampled` or `frac_num_sampled`.

        .. note::
            You must give `frac_mass_sampled` *or* `frac_num_sampled`.

        Parameters
        ----------
        imf : str
            Name for the IMF.
        bandpass : str
            Observation bandpass of the limiting magnitude.
        frac_mass_sampled: float, optional
            Fraction of mass that will be sampled if only stars that are
            brighter than the magnitude limit are sampled.
        frac_num_sampled: float, optional
            Fraction of stars by number that will be sampled if only stars that
            are brighter than the magnitude limit are sampled.
        distance : float or `~astropy.units.Quantity`, optional
            Distance to source. If float is given, the units are assumed
            to be `~astropy.units.Mpc`. Default distance is 10
            `~astropy.units.pc` (i.e., the mags are in absolute units).

        Returns
        -------
        mag_limit : float
            The desired limiting magnitude.
        mass_limit : float
            Mass of stars (in solar masses) that have the limiting magnitude.
        """
        mfint = IMFIntegrator(imf, self.m_min, self.m_max)
        m_vals = np.logspace(np.log10(self.m_min), np.log10(self.m_max), 500)
        err_msg = 'You must give a fraction such that 0 < fraction <= 1.'
        if frac_mass_sampled is not None:
            assert frac_mass_sampled > 0 and frac_mass_sampled <= 1, err_msg
            frac_interp = frac_mass_sampled
            fracs = [mfint.m_integrate(_m, self.m_max, True) for _m in m_vals]
        elif frac_num_sampled is not None:
            assert frac_num_sampled > 0 and frac_num_sampled <= 1, err_msg
            frac_interp = frac_num_sampled
            fracs = [mfint.integrate(_m, self.m_max, True) for _m in m_vals]
        else:
            msg = 'You must give either frac_mass_sampled or frac_num_sampled.'
            raise Exception(msg)
        d = check_units(distance, 'Mpc')
        mass_limit = interp1d(fracs, m_vals)([frac_interp])[0]
        mag_limit = self.interpolate(bandpass, [mass_limit])[0]
        mag_limit = mag_limit + 5 * np.log10(d.to('pc').value) - 5
        return mag_limit, mass_limit

    def nearest_mini(self, m):
        """
        Find the nearest mass to `m` in the initial mass array.

        Parameters
        ----------
        m : float:
            The mass for which we want the nearest initial mass.

        Returns
        -------
        m_nearest : float
            Value of the nearest mass.
        arg_nearest : int
            Argument of the nearest mass.
        """
        arg_nearest = np.abs(self.mini - m).argmin()
        m_nearest = self.mini[arg_nearest]
        return m_nearest, arg_nearest

    def imf_weights(self, imf, m_min_norm=None, m_max_norm=None,
                    norm_type='mass', **kwargs):
        """
        Calculate IMF weights.

        Parameters
        ----------
        imf : str or callable function that takes mass as its argument
            Stellar Initial mass function.
        m_min_norm : None or float, optional
            Minimum mass for the normalization. Must be less than or equal to
            the mini mass of isochrone, which will be used if None is given.
        m_max_norm : float, optional
            Maximum mass for normalization.
        norm_type : str, optional
            Type of IMF normalization (mass or number).

        Returns
        -------
        wght : `~numpy.ndarray`
            IMF weights calculated such that an integral over mass is simply
            given by SUM(m * wght).
        """
        m_min = m_min_norm if m_min_norm else self.mini.min()
        m_max = m_max_norm if m_max_norm else self.mini.max()
        if m_min > self.m_min:
            raise Exception('Minimum mass must be <= isochrone min mass.')
        if callable(imf):
            imf_func = imf
        else:
            imf_func = lambda m: imf_dict[imf](m, norm_type=None)
        m_imf_func = lambda m: m * imf_func(m)
        norm_func = dict(mass=m_imf_func, number=imf_func)[norm_type]
        norm = quad(norm_func, m_min, m_max, **kwargs)[0]

        wght = []
        mini = self.mini
        # Assume mass is constant in each bin and integrate.
        # This means an integral over mass is simply SUM(m * wght)
        # This trick was stolen from Charlie Conroy's FSPS code :)
        for i in range(len(mini)):
            if i == 0:
                m1 = m_min
            else:
                m1 = mini[i] - 0.5 * (mini[i] - mini[i-1])
            if i == len(mini) - 1:
                m2 = mini[i]
            else:
                m2 = mini[i] + 0.5 * (mini[i+1] - mini[i])
            if m2 < m1:
                raise Exception('Masses must be monotonically increasing.')
            wght.append(quad(imf_func, m1, m2, **kwargs)[0])
        wght = np.array(wght) / norm
        return wght

    def ssp_color(self, blue, red, imf='kroupa', **kwargs):
        """
        Calculate IMF-weighted integrated color.

        Parameters
        ----------
        blue : str
            Blue bandpass.
        red : str
            Red bandpass.
        imf : str, optional
            IMF name.

        Returns
        -------
        color : float
            Integrated color of stellar population.
        """
        wght = self.imf_weights(imf, **kwargs)
        lum_blue = np.sum(wght * 10**(-0.4 * self.mag_table[blue]))
        lum_red = np.sum(wght * 10**(-0.4 * self.mag_table[red]))
        color = -2.5 * np.log10(lum_blue / lum_red)
        return color

    def ssp_sbf_mag(self, bandpass, imf='kroupa', **kwargs):
        """
        Calculate IMF-weighted SBF magnitude.

        Parameters
        ----------
        bandpass : str
            Bandpass to of SBF mag.
        imf : str, optional
            IMF name.

        Returns
        -------
        sbf : float
            SBF magnitude of stellar population.
        """
        wght = self.imf_weights(imf, **kwargs)
        lumlum = np.sum(wght * 10**(-0.8 * self.mag_table[bandpass]))
        lum = np.sum(wght * 10**(-0.4 * self.mag_table[bandpass]))
        sbf = -2.5 * np.log10(lumlum / lum)
        return sbf

    def ssp_mag(self, bandpass, imf='kroupa', norm_type='mass', **kwargs):
        """
        Calculate IMF-weighted magnitude.

        Parameters
        ----------
        bandpass : str
            Bandpass to of mean mag.
        imf : str, optional
            IMF name.
        norm_type : str, optional
            Normalization type (mass or number)

        Returns
        -------
        m : float
            Integrated magnitude of stellar population.
        """
        wght = self.imf_weights(imf, norm_type=norm_type, **kwargs)
        lum = np.sum(wght * 10**(-0.4 * self.mag_table[bandpass]))
        m = -2.5 * np.log10(lum)
        return m

    def ssp_mean_star_mass(self, imf):
        """
        Calculate IMF-weighted mean stellar mass, where the IMF is normalized
        by number to one star formed. The mean stellar mass is equivalent to
        dividing a large of samples (distributed according to the IMF) by the
        number of stars.

        Parameters
        ----------
        imf : str:
            Initial mass function name.

        Returns
        -------
        mean_mass : float
            The mean stellar mass.
        """
        w = self.imf_weights(imf, norm_type='number',
                             m_max_norm=self.mini.max())
        mean_mass = np.sum(self.mact * w)
        return mean_mass

    def ssp_surviving_mass(self, imf, m_min=None, m_max=120,
                           add_remnants=True, mlim_bh=40.0, mlim_ns=8.5):
        """
        Calculate IMF-weighted mass normalized such that 1 solar mass is
        formed by the age of the population.

        The initial-mass-dependent remnant formulae are taken
        from `Renzini & Ciotti 1993
        <https://ui.adsabs.harvard.edu/abs/1993ApJ...416L..49R/abstract>`_.

        Parameters
        ----------
        imf : str
            IMF name.
        add_remnants : bool
            If True, add stellar remnants to surviving mass.

        Returns
        -------
        mass : float
            Total surviving mass.
        """
        m_min = m_min if m_min else self.mini.min()
        m_max = m_max if m_max else self.mini.max()
        wght = self.imf_weights(imf, m_min_norm=m_min, m_max_norm=m_max)
        mass = np.sum(self.mact * wght)

        if add_remnants:
            imf_func = lambda m: imf_dict[imf](m, norm_type=None)
            m_imf_func = lambda m: m * imf_func(m)
            norm = quad(m_imf_func, m_min, m_max)[0]

            # BH remnants
            m_low = max(mlim_bh, self.m_max)
            mass = mass + 0.5 * quad(m_imf_func, m_low, m_max)[0] / norm

            # NS remnants
            if self.m_max <= mlim_bh:
                m_low = max(mlim_ns, self.m_max)
                mass = mass + 1.4 * quad(imf_func, m_low, mlim_bh)[0] / norm

            # WD remnants
            if self.m_max < mlim_ns:
                mmax = self.m_max
                mass = mass + 0.48 * quad(imf_func, mmax, mlim_ns)[0] / norm
                mass = mass + 0.077 * quad(m_imf_func, mmax, mlim_ns)[0] / norm

        return mass


def fetch_mist_iso_cmd(log_age, feh, phot_system, mist_path=MIST_PATH,
                       v_over_vcrit=0.4):
    """
    Fetch MIST isochrone grid.

    Parameters
    ----------
    log_age : float
        Logarithm base 10 of the simple stellar population age in years.
    feh : float
        Metallicity [Fe/H] of the simple stellar population.
    phot_system : str
        Name of the photometric system.
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the `MIST_PATH` environment variable.
    v_over_vcrit : float, optional
        Rotation rate divided by the critical surface linear velocity. Current
        options are 0.4 (default) and 0.0.

    Returns
    -------
    iso_cmd : `~numpy.ndarray`
        Structured ``numpy`` array with isochrones and stellar magnitudes.
    """

    # fetch the mist grid if necessary
    fetch_mist_grid_if_needed(phot_system, v_over_vcrit, mist_path)

    v = f'{v_over_vcrit:.1f}'
    ver = 'v1.2'
    p = phot_str_helper[phot_system.lower()]
    path = os.path.join(mist_path, 'MIST_' + ver + f'_vvcrit{v}_' + p)
    sign = 'm' if feh < 0 else 'p'
    fn = f'MIST_{ver}_feh_{sign}{abs(feh):.2f}_afe_p0.0_vvcrit{v}_{p}.iso.cmd'
    fn = os.path.join(path, fn)
    iso_cmd = IsoCmdReader(fn, verbose=False)
    iso_cmd = iso_cmd.isocmds[iso_cmd.age_index(log_age)]
    return iso_cmd


class MISTIsochrone(Isochrone):
    """
    Class for fetching and storing MIST isochrones. It also has several methods
    for calculating IMF-weighted photometric properties of a stellar population
    with the given age an metallicity.

    .. note::
        Currently, the models are interpolated in metallicity but not in age.
        Ages are therefore limited to the age grid of the MIST models. The
        [Fe/H] and log(Age/yr) grids are stored as the private class attributes
        ``_feh_grid`` and ``_log_age_grid``.

    Parameters
    ----------
    log_age : float
        Logarithm base 10 of the simple stellar population age in years.
    feh : float
        Metallicity [Fe/H] of the simple stellar population.
    phot_system : str or list-like
        Name of the photometric system(s).
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the ``MIST_PATH`` environment variable.
    ab_or_vega : str, optional
        Magnitudes will be in the AB (default) or Vega magnitude system.
    v_over_vcrit : float, optional
        Rotation rate divided by the critical surface linear velocity. Current
        options are 0.4 (default) and 0.0.
    """

    # the age grid
    _log_age_grid = np.arange(5.0, 10.3, 0.05)
    _log_age_min = _log_age_grid.min()
    _log_age_max = _log_age_grid.max()

    # the [Fe/H] metallicity grid
    # mist has feh <= -4, but using <=-3 due to interpolation issues
    _feh_grid = np.concatenate([np.arange(-3.0, -2., 0.5),
                                np.arange(-2.0, 0.75, 0.25)])
    _feh_min = _feh_grid.min()
    _feh_max = _feh_grid.max()

    def __init__(self, log_age, feh, phot_system, mist_path=MIST_PATH,
                 ab_or_vega='ab', v_over_vcrit=0.4):

        # verify age are metallicity are within model grids
        if log_age < self._log_age_min or log_age > self._log_age_max:
            raise Exception(f'log_age = {log_age} not in range of age grid')
        if feh < self._feh_min or feh > self._feh_max:
            raise Exception(f'feh = {feh} not in range of feh grid')

        self.feh = feh
        self.mist_path = mist_path
        self.phot_system = phot_system
        self.ab_or_vega = ab_or_vega
        self.v_over_vcrit = v_over_vcrit

        # use nearest age (currently not interpolating on age)
        age_diff = np.abs(self._log_age_grid - log_age)
        self.log_age = self._log_age_grid[age_diff.argmin()]
        if age_diff.min() > 1e-6:
            logger.debug('Using nearest log_age = {:.2f}'.format(self.log_age))

        # store phot_system as list to allow multiple photometric systems
        if type(phot_system) == str:
            phot_system = [phot_system]

        # fetch first isochrone grid, interpolating on [Fe/H] if necessary
        self._iso_full = self._fetch_iso(phot_system[0])

        # iterate over photometric systems and fetch remaining isochrones
        filter_dict = get_filter_names()
        filters = filter_dict[phot_system[0]].copy()
        for p in phot_system[1:]:
            filt = filter_dict[p].copy()
            filters.extend(filt)
            _iso = self._fetch_iso(p)
            mags = [_iso[f].data for f in filt]
            self._iso_full = append_fields(self._iso_full, filt, mags)

        # covert magnitudes to ab or vega as necessary
        self.zpt_convert = load_zero_point_converter()
        for filt in filters:
            converter = getattr(self.zpt_convert, f'to_{ab_or_vega.lower()}')
            try:
                m_convert = converter(filt)
            except AttributeError:
                m_convert = 0.0
                logger.warning(f'No AB / Vega conversion found for {filt}.')
            self._iso_full[filt] = self._iso_full[filt] + m_convert

        super(MISTIsochrone, self).__init__(
            mini = self._iso_full['initial_mass'],
            mact = self._iso_full['star_mass'],
            mags = Table(self._iso_full[filters]),
            eep = self._iso_full['EEP'],
            log_L = self._iso_full['log_L'],
            log_Teff = self._iso_full['log_Teff'],
        )

    @property
    def isochrone_full(self):
        """MIST entire isochrone in a structured `~numpy.ndarray`."""
        return self._iso_full

    @staticmethod
    def from_parsec(fn, **kwargs):
        msg = 'PARSEC isochrones do not with MISTIsochrone.'
        raise Exception(msg + ' Use artpop.Isochrone instead.')

    def _fetch_iso(self, phot_system):
        """Fetch MIST isochrone grid, interpolating on [Fe/H] if necessary."""
        if self.feh in self._feh_grid:
            args = [self.log_age, self.feh, phot_system, self.mist_path,
                    self.v_over_vcrit]
            iso = fetch_mist_iso_cmd(*args)
        else:
            iso = self._interp_on_feh(phot_system)
        return iso

    def _interp_on_feh(self, phot_system):
        """Interpolate isochrones between two [Fe/H] grid points."""
        i_feh = self._feh_grid.searchsorted(self.feh)
        feh_lo, feh_hi = self._feh_grid[i_feh - 1: i_feh + 1]

        logger.debug('Interpolating to [Fe/H] = {:.2f} '\
                     'using [Fe/H] = {} and {}'.\
                     format(self.feh, feh_lo, feh_hi))

        mist_0 = fetch_mist_iso_cmd(
            self.log_age, feh_lo, phot_system, self.mist_path)
        mist_1 = fetch_mist_iso_cmd(
            self.log_age, feh_hi, phot_system, self.mist_path)

        y0, y1 = np.array(mist_0.tolist()), np.array(mist_1.tolist())

        x = self.feh
        x0, x1 = feh_lo, feh_hi
        weight = (x - x0) / (x1 - x0)

        len_0, len_1 = len(y0), len(y1)

        # if necessary, extrapolate using trend of the longer array
        if (len_0 < len_1):
            delta = y1[len_0:] - y1[len_0 - 1]
            y0 = np.append(y0, y0[-1] + delta, axis=0)
        elif (len_0 > len_1):
            delta = y0[len_1:] - y0[len_1 - 1]
            y1 = np.append(y1, y1[-1] + delta, axis=0)

        y = y0 * (1 - weight) + y1 * weight
        iso = np.core.records.fromarrays(y.transpose(), dtype=mist_0.dtype)

        return iso



import os, sys
import builtins
from setuptools import setup, find_packages
sys.path.append('src')


def readme():
    with open("README.rst") as f:
        return f.read()


# HACK: fetch version
builtins.__ARTPOP_SETUP__ = True
import artpop
version = artpop.__version__


# Publish the library to PyPI.
if "publish" in sys.argv[-1]:
    os.system("python setup.py sdist bdist_wheel")
    os.system(f"python3 -m twine upload dist/*{version}*")
    sys.exit()


# Push a new tag to GitHub.
if "tag" in sys.argv:
    os.system("git tag -a {0} -m 'version {0}'".format(version))
    os.system("git push --tags")
    sys.exit()


setup(
    name='artpop',
    version=version,
    description='Building artificial galaxies one star at a time',
    long_description=readme(),
    author='Johnny Greco & Shany Danieli',
    author_email='artpopcode@gmail.com',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"artpop": ["data/*.pkl", "data/*.txt"]},
    include_package_data=True,
    url='https://github.com/ArtificialStellarPopulations/ArtPop',
    install_requires=[
        'numpy>=1.17',
        'scipy>=1',
        'astropy>=4',
        'matplotlib>=3',
        'fast-histogram>=0.9',
        'requests',
     ],
     classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Astronomy",
      ],
    python_requires='>=3.6',
)



# Standard library
from copy import deepcopy

# Third-party
import numpy as np
from astropy import units as u
from astropy.table import Table, vstack
from astropy.modeling.models import Sersic2D
from scipy.special import gammaincinv, gamma

# Project
from . import MIST_PATH
from .stars import SSP, MISTSSP, MISTIsochrone, constant_sb_stars_per_pix
from .space import sersic_xy, plummer_xy, uniform_xy, Plummer2D, Constant2D
from .filters import load_zero_point_converter
from .util import check_units, check_xy_dim


__all__ = [
    'Source',
    'SersicSP',
    'MISTSersicSSP',
    'PlummerSP',
    'MISTPlummerSSP',
    'UniformSSP',
    'MISTUniformSSP',
]


class Source(object):
    """
    Artificial stellar system to be mock imaged.

    Parameters
    ----------
    xy : `~numpy.ndarray` or `~numpy.ma.MaskedArray`
        Positions of the stars in (x, y) coordinates.
    mags : dict of `~numpy.ndarray` or `~astropy.table.Table`
        Stellar magnitudes.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`, optional
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`.
    labels : list-like, optional
        Labels for the stars. For example, EEP values (int or float) or name
        of evolutionary phase (str).
    """
    def __init__(self, xy, mags, xy_dim, pixel_scale=None, labels=None):
        if type(mags) == dict:
            self.mags = Table(mags)
        else:
            self.mags = mags.copy()
        if len(xy) != len(self.mags):
            raise Exception('numbers of magnitudes and positions must match')
        if type(xy) == np.ma.MaskedArray:
            mask = np.any(~xy.mask, axis=1)
            self.mags = self.mags[mask]
            self.xy = xy.data[mask]
        else:
            self.xy = np.asarray(xy)
            mask = np.ones(len(self.xy), dtype=bool)
        self.labels = labels if labels is None else np.asarray(labels)[mask]
        self.xy_dim = check_xy_dim(xy_dim)
        if pixel_scale is not None:
            self.pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)

    @property
    def x(self):
        """x coordinates"""
        return self.xy[:, 0]

    @property
    def y(self):
        """y coordinates"""
        return self.xy[:, 1]

    @property
    def num_stars(self):
        """Number of stars in source"""
        return len(self.x)

    def copy(self):
        """Create deep copy of source object."""
        return deepcopy(self)

    def select_stars(self, label):
        """Return boolean mask that is set to True for stars of given label."""
        if self.labels is None:
            raise Exception('The stars of this source do not have labels.')
        return self.labels == label

    def __add__(self, src):
        if Source not in src.__class__.__mro__:
            raise Exception(f'{type(src)} is not a valid Source object')
        new = deepcopy(self)
        if not np.allclose(new.xy_dim, src.xy_dim):
            raise Exception('Sources must have the same xy_dim')
        if new.labels is None:
            new.labels = np.ones(new.num_stars, dtype=int)
        labels = np.ones(src.num_stars, dtype=int) * (new.labels.max() + 1)
        new_labels = np.concatenate([new.labels, labels])
        new.xy = np.concatenate([new.xy, src.xy])
        new.mags = vstack([new.mags, src.mags])
        comp_src = CompositeSource(new.xy, new.mags, new.xy_dim,
                                   new.pixel_scale, new_labels)
        return comp_src


class CompositeSource(Source):
    """A source consisting of 2 or more `~artpop.source.Source` objects."""
    pass


class SersicSP(Source):
    """
    Stellar population with a Sersic spatial distribution.

    Parameters
    ----------
    sp : `~artpop.stars.populations.StellarPopulation`
        A stellar population object.
    r_eff : float or `~astropy.units.Quantity`
        Effective radius of the source. If a float is given, the units are
        assumed to be `kpc`. Must be greater than zero.
    n : float
        Sersic index. Must be greater than zero.
    theta : float or `~astropy.units.Quantity`
        Rotation angle, counterclockwise from the positive x-axis. If a float
        is given, the units are assumed to be `degree`.
    ellip : float
        Ellipticity.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    num_r_eff : float, optional
        Number of r_eff to sample positions within. This parameter is needed
        because the current Sersic sampling function samples from within a
        discrete grid. Default is 10.
    dx : float, optional
        Shift from center of image in the x direction.
    dy : float, optional
        Shift from center of image in the y direction.
    labels : list-like, optional
        Labels for the stars. For example, EEP values (int or float) or name
        of evolutionary phase (str).
    """

    def __init__(self, sp, r_eff, n, theta, ellip, xy_dim, pixel_scale,
                 num_r_eff=10, dx=0, dy=0, labels=None):

        self.sp = sp
        self.mag_limit = sp.mag_limit
        self.mag_limit_band = sp.mag_limit_band
        self.smooth_model = None

        if self.mag_limit is not None and sp.frac_num_sampled < 1.0:
            _r_eff = check_units(r_eff, 'kpc').to('Mpc').value
            _theta = check_units(theta, 'deg').to('radian').value
            _distance = check_units(sp.distance, 'Mpc').to('Mpc').value
            _pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)

            if _r_eff <= 0:
                raise Exception('Effective radius must be greater than zero.')

            xy_dim = check_xy_dim(xy_dim)
            x_0, y_0 = xy_dim//2
            x_0 += dx
            y_0 += dy
            self.n = n
            self.ellip = ellip
            self.r_sky = np.arctan2(_r_eff, _distance) * u.radian.to('arcsec')
            self.r_sky *= u.arcsec
            r_pix = self.r_sky.to('pixel', u.pixel_scale(_pixel_scale)).value

            self.smooth_model = Sersic2D(
                x_0=x_0, y_0=y_0, n=n, r_eff=r_pix, theta=_theta, ellip=ellip)

        self.xy_kw = dict(num_stars=sp.num_stars, r_eff=r_eff, n=n,
                          theta=theta, ellip=ellip, distance=sp.distance,
                          xy_dim=xy_dim, num_r_eff=num_r_eff, dx=dx, dy=dy,
                          pixel_scale=pixel_scale, random_state=sp.rng)

        _xy = sersic_xy(**self.xy_kw)

        super(SersicSP, self).__init__(
            _xy, sp.mag_table, xy_dim, pixel_scale, labels)

    def mag_to_image_amplitude(self, m_tot, zpt):
        """
        Convert total magnitude into amplitude parameter for the smooth model.

        Parameters
        ----------
        m_tot : float
            Total magnitude in the smooth component of the system.
        zpt : float
            Photometric zero point.

        Returns
        -------
        mu_e : float
            Surface brightness at the effective radius of the Sersic
            distribution in mags per square arcsec.
        amplitude : float
            Amplitude parameter for the smooth model in image flux units.
        param_name : str
            Name of amplitude parameter (needed to set its value when
            generating the smooth model).
        """
        param_name = 'amplitude'
        b_n = gammaincinv(2.0 * self.n, 0.5)
        f_n = gamma(2 * self.n) *self.n * np.exp(b_n) / b_n**(2 * self.n)
        r_circ = self.r_sky * np.sqrt(1 - self.ellip)
        area = np.pi * r_circ.to('arcsec').value**2
        mu_e = m_tot + 2.5 * np.log10(2 * area) + 2.5 * np.log10(f_n)
        amplitude = 10**(0.4 * (zpt - mu_e)) * self.pixel_scale.value**2
        return mu_e, amplitude, param_name


def _check_label_type(ssp, label_type, has_phases=True):
    if label_type is not None:
        if label_type == 'phases' and has_phases:
            labels = ssp.get_star_phases()
        elif hasattr(ssp, label_type):
            labels = getattr(ssp, label_type)
        else:
            raise Exception(f'{label_type} is not a valid label type.')
    else:
        labels = None
    return labels


class MISTSersicSSP(SersicSP):
    """
    MIST simple stellar population with a Sersic spatial distribution. This
    is a convenience class that combines `~artpop.space.sersic_xy` and
    `~artpop.stars.MISTSSP` to make a `~artpop.source.Source` object.

    .. note::
        You must give `total_mass` *or* `num_stars`.

    Parameters
    ----------
    log_age : float
        Log (base 10) of the simple stellar population age in years.
    feh : float
        Metallicity [Fe/H] of the simple stellar population.
    phot_system : str or list-like
        Name of the photometric system(s).
    r_eff : float or `~astropy.units.Quantity`
        Effective radius of the source. If a float is given, the units are
        assumed to be `kpc`. Must be greater than zero.
    n : float
        Sersic index. Must be greater than zero.
    theta : float or `~astropy.units.Quantity`
        Rotation angle, counterclockwise from the positive x-axis. If a float
        is given, the units are assumed to be `degree`.
    ellip : float
        Ellipticity defined as `1 - b/a`, where `b` is the semi-minor axis
        and `a` is the semi-major axis.
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `Mpc`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    num_stars : int or `None`
        Number of stars in source. If `None`, then must give `total_mass`.
    total_mass : float or `None`
        Stellar mass of the source. If `None`, then must give `num_stars`. This
        mass accounts for stellar remnants, so the actual sampled mass will be
        less than the given value.
    add_remnants : bool, optional
        If True (default), apply scaling factor to total mass to account for
        stellar remnants in the form of white dwarfs, neutron stars,
        and black holes.
    mag_limit : float, optional
        Only sample individual stars that are brighter than this magnitude. All
        fainter stars will be combined into an integrated component. Otherwise,
        all stars in the population will be sampled. You must also give the
        `mag_limit_band` if you use this parameter.
    mag_limit_band : str, optional
        Bandpass of the limiting magnitude. You must give this parameter if
        you use the `mag_limit` parameter.
    imf : str, optional
        The initial stellar mass function. Default is `'kroupa'`.
    imf_kw : dict, optional
        Optional keyword arguments for sampling the stellar mass function.
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the `MIST_PATH` environment variable.
    num_r_eff : float, optional
        Number of r_eff to sample positions within. This parameter is needed
        because the current Sersic sampling function samples from within a
        discrete grid. Default is 10.
    mass_tolerance : float, optional
        Tolerance in the fractional difference between the input mass and the
        final mass of the population. The parameter is only used when
        `total_mass` is given.
    dx : float, optional
        Shift from center of image in the x direction.
    dy : float, optional
        Shift from center of image in the y direction.
    label_type : str, optional
        If not None, the type of labels to pass to the `~artpop.source.Source`
        object. Can be `phases` or an attribute of `~artpop.stars.MISTSSP`.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    """

    def __init__(self, log_age, feh, phot_system, r_eff, n, theta, ellip,
                 distance, xy_dim, pixel_scale, num_stars=None,
                 total_mass=None, add_remnants=True, mag_limit=None,
                 mag_limit_band=None, imf='kroupa', imf_kw={},
                 mist_path=MIST_PATH, num_r_eff=10, mass_tolerance=0.01,
                 dx=0, dy=0, label_type=None, random_state=None):

        self.ssp_kw = dict(log_age=log_age, feh=feh, phot_system=phot_system,
                           distance=distance, total_mass=total_mass,
                           num_stars=num_stars, imf=imf, mist_path=mist_path,
                           imf_kw=imf_kw, random_state=random_state,
                           mag_limit=mag_limit, mag_limit_band=mag_limit_band,
                           add_remnants=add_remnants,
                           mass_tolerance=mass_tolerance)

        ssp = MISTSSP(**self.ssp_kw)
        labels = _check_label_type(ssp, label_type)

        super(MISTSersicSSP, self).__init__(
            sp=ssp, r_eff=r_eff, n=n, theta=theta, ellip=ellip, xy_dim=xy_dim,
            pixel_scale=pixel_scale, num_r_eff=num_r_eff, dx=dx, dy=dy,
            labels=labels)


class PlummerSP(Source):
    """
    Simple stellar population with a Plummer spatial distribution.

    Parameters
    ----------
    sp : `~artpop.stars.populations.StellarPopulation`
        A stellar population object.
    scale_radius : float or `~astropy.units.Quantity`
        Scale radius of the source. If a float is given, the units are
        assumed to be `kpc`. Must be greater than zero.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    dx : float, optional
        Shift from center of image in the x direction.
    dy : float, optional
        Shift from center of image in the y direction.
    labels : list-like, optional
        Labels for the stars. For example, EEP values (int or float) or name
        of evolutionary phase (str).
    """

    def __init__(self, sp, scale_radius, xy_dim, pixel_scale, dx=0, dy=0,
                 labels=None):

        self.sp = sp
        self.mag_limit = sp.mag_limit
        self.mag_limit_band = sp.mag_limit_band
        self.smooth_model = None

        if self.mag_limit is not None and sp.frac_num_sampled < 1.0:
            _rs = check_units(scale_radius, 'kpc').to('Mpc').value
            _distance = check_units(sp.distance, 'Mpc').to('Mpc').value
            _pixel_scale = check_units(pixel_scale, u.arcsec / u.pixel)

            self.r_sky = np.arctan2(_rs, _distance) * u.radian.to('arcsec')
            self.r_sky *= u.arcsec
            r_pix = self.r_sky.to('pixel', u.pixel_scale(_pixel_scale)).value

            xy_dim = check_xy_dim(xy_dim)
            x_0, y_0 = xy_dim//2
            x_0 += dx
            y_0 += dy

            self.smooth_model = Plummer2D(x_0=x_0, y_0=y_0, scale_radius=r_pix)

        self.xy_kw = dict(num_stars=sp.num_stars, scale_radius=scale_radius,
                          distance=sp.distance, xy_dim=xy_dim,
                          pixel_scale=pixel_scale, dx=dx, dy=dy,
                          random_state=sp.rng)

        _xy = plummer_xy(**self.xy_kw)
        super(PlummerSP, self).__init__(
            _xy, sp.mag_table, xy_dim, pixel_scale, labels)

    def mag_to_image_amplitude(self, m_tot, zpt):
        """
        Convert total magnitude into amplitude parameter for the smooth model.

        Parameters
        ----------
        m_tot : float
            Total magnitude in the smooth component of the system.
        zpt : float
            Photometric zero point.

        Returns
        -------
        mu_0 : float
            Central surface brightness of the Plummer distribution in
            mags per square arcsec.
        amplitude : float
            Amplitude parameter for the smooth model in image flux units.
        param_name : str
            Name of amplitude parameter (needed to set its value when
            generating the smooth model).
        """
        param_name = 'amplitude'
        area = np.pi * self.r_sky.to('arcsec').value**2
        flux = 10**(0.4 * (zpt - m_tot))
        mu_0 = zpt - 2.5 * np.log10(flux / area)
        amplitude = (flux / area) * self.pixel_scale.value**2
        return mu_0, amplitude, param_name


class MISTPlummerSSP(PlummerSP):
    """
    MIST simple stellar population with a Plummer spatial distribution.

    .. note::
        You must give `total_mass` *or* `num_stars`.

    Parameters
    ----------
    log_age : float
        Log (base 10) of the simple stellar population age in years.
    feh : float
        Metallicity [Fe/H] of the simple stellar population.
    phot_system : str or list-like
        Name of the photometric system(s).
    scale_radius : float or `~astropy.units.Quantity`
        Scale radius of the source. If a float is given, the units are
        assumed to be `kpc`. Must be greater than zero.
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `Mpc`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    num_stars : int or `None`
        Number of stars in source. If `None`, then must give `total_mass`.
    total_mass : float or `None`
        Stellar mass of the source. If `None`, then must give `num_stars`. This
        mass accounts for stellar remnants, so the actual sampled mass will be
        less than the given value.
    add_remnants : bool, optional
        If True (default), apply scaling factor to total mass to account for
        stellar remnants in the form of white dwarfs, neutron stars,
        and black holes.
    mag_limit : float, optional
        Only sample individual stars that are brighter than this magnitude. All
        fainter stars will be combined into an integrated component. Otherwise,
        all stars in the population will be sampled. You must also give the
        `mag_limit_band` if you use this parameter.
    mag_limit_band : str, optional
        Bandpass of the limiting magnitude. You must give this parameter if
        you use the `mag_limit` parameter.
    imf : str, optional
        The initial stellar mass function. Default is `'kroupa'`.
    imf_kw : dict, optional
        Optional keyword arguments for sampling the stellar mass function.
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the `MIST_PATH` environment variable.
    mass_tolerance : float, optional
        Tolerance in the fractional difference between the input mass and the
        final mass of the population. The parameter is only used when
        `total_mass` is given.
    dx : float, optional
        Shift from center of image in the x direction.
    dy : float, optional
        Shift from center of image in the y direction.
    label_type : str, optional
        If not None, the type of labels to pass to the `~artpop.source.Source`
        object. Can be `phases` or an attribute of `~artpop.stars.MISTSSP`.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    """

    def __init__(self, log_age, feh, phot_system, scale_radius,
                 distance, xy_dim, pixel_scale, num_stars=None,
                 total_mass=None, add_remnants=True, mag_limit=None,
                 mag_limit_band=None, imf='kroupa', imf_kw={},
                 mist_path=MIST_PATH, mass_tolerance=0.01, dx=0, dy=0,
                 label_type=None, random_state=None):

        self.ssp_kw = dict(log_age=log_age, feh=feh, phot_system=phot_system,
                           distance=distance, total_mass=total_mass,
                           num_stars=num_stars, imf=imf, mist_path=mist_path,
                           imf_kw=imf_kw, random_state=random_state,
                           mag_limit=mag_limit, mag_limit_band=mag_limit_band,
                           add_remnants=add_remnants,
                           mass_tolerance=mass_tolerance)

        ssp = MISTSSP(**self.ssp_kw)
        labels = _check_label_type(ssp, label_type)

        super(MISTPlummerSSP, self).__init__(
            sp=ssp, scale_radius=scale_radius, xy_dim=xy_dim,
            pixel_scale=pixel_scale, dx=dx, dy=dy, labels=labels)


class UniformSSP(Source):
    """
    Simple stellar population with a uniform spatial distribution.

    Parameters
    ----------
    isochrone  : `~artpop.stars.Isochrone`
        Isochrone object.
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `Mpc`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    sb : float
        Surface brightness of of uniform stellar distribution.
    sb_band : str
        Photometric filter of ``sb``. Must be part of ``phot_system``.
    mag_limit : float, optional
        Only sample individual stars that are brighter than this magnitude. All
        fainter stars will be combined into an integrated component. Otherwise,
        all stars in the population will be sampled. You must also give the
        `mag_limit_band` if you use this parameter.
    mag_limit_band : str, optional
        Bandpass of the limiting magnitude. You must give this parameter if
        you use the `mag_limit` parameter.
    pixel_scale : float or `~astropy.units.Quantity`, optional
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    imf : str, optional
        The initial stellar mass function. Default is `'kroupa'`.
    imf_kw : dict, optional
        Optional keyword arguments for sampling the stellar mass function.
    ssp_kw : dict, optional
        Additional keyword arguments for `~artpop.stars.SSP`.
    label_type : str, optional
        If not None, the type of labels to pass to the `~artpop.source.Source`
        object. Must be an attribute of `~artpop.stars.SSP`.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    """

    def __init__(self, isochrone, distance, xy_dim, pixel_scale, sb, sb_band,
                 mag_limit=None, mag_limit_band=None, imf='kroupa', imf_kw={},
                 ssp_kw={}, label_type=None, random_state=None):

        self.isochrone = isochrone
        self.mag_limit = mag_limit
        self.mag_limit_band = mag_limit_band
        self.smooth_model = None

        mean_mag = isochrone.ssp_mag(
            sb_band, norm_type='number', m_min_norm=isochrone.m_min,
            m_max_norm=isochrone.m_max)

        stars_per_pix = constant_sb_stars_per_pix(
            sb, mean_mag, distance, pixel_scale)

        xy_dim  = check_xy_dim(xy_dim)
        self.ssp_kw = ssp_kw
        self.ssp_kw['mag_limit'] = mag_limit
        self.ssp_kw['mag_limit_band'] = mag_limit_band
        self.ssp_kw['distance'] = distance
        self.ssp_kw['total_mass'] = None
        self.ssp_kw['num_stars'] = int(stars_per_pix * np.multiply(*xy_dim))

        if type(isochrone) == MISTIsochrone:
            self.ssp_kw['log_age'] = isochrone.log_age
            self.ssp_kw['feh'] = isochrone.feh
            self.ssp_kw['phot_system'] = isochrone.phot_system
            self.ssp_kw['v_over_vcrit'] = isochrone.v_over_vcrit
            self.ssp_kw['mist_path'] = isochrone.mist_path
            self.sp = MISTSSP(**self.ssp_kw)
            labels = _check_label_type(self.sp, label_type)
        else:
            self.sp = SSP(isochrone, **self.ssp_kw)
            labels = _check_label_type(self.sp, label_type, False)

        if mag_limit is not None and self.sp.frac_num_sampled < 1.0:
            self.smooth_model = Constant2D()

        self.xy_kw = dict(num_stars=self.sp.num_stars,
                          xy_dim=xy_dim, random_state=self.sp.rng)

        _xy = uniform_xy(**self.xy_kw)
        super(UniformSSP, self).__init__(
            _xy, self.sp.mag_table, xy_dim, pixel_scale, labels)

    def mag_to_image_amplitude(self, m_tot, zpt):
        """
        Convert total magnitude into amplitude parameter for the smooth model.

        Parameters
        ----------
        m_tot : float
            Total magnitude in the smooth component of the system.
        zpt : float
            Photometric zero point.

        Returns
        -------
        mu : float
            Constant surface brightness in mags per square arcsec.
        amplitude : float
            Amplitude parameter for the smooth model in image flux units.
        param_name : str
            Name of amplitude parameter (needed to set its value when
            generating the smooth model).
        """
        param_name = 'amplitude'
        area = np.multiply(*self.xy_dim)
        flux = 10**(0.4 * (zpt - m_tot))
        amplitude = flux / area
        mu = zpt - 2.5 * np.log10(amplitude / self.pixel_scale.value**2)
        return mu, amplitude, param_name


class MISTUniformSSP(UniformSSP):
    """
    MIST simple stellar population with a uniform spatial distribution.

    .. note::
        You must give `total_mass` *or* `num_stars`.

    Parameters
    ----------
    log_age : float
        Log (base 10) of the simple stellar population age in years.
    feh : float
        Metallicity [Fe/H] of the simple stellar population.
    phot_system : str or list-like
        Name of the photometric system(s).
    distance : float or `~astropy.units.Quantity`
        Distance to source. If float is given, the units are assumed
        to be `Mpc`.
    xy_dim : int or list-like
        Dimensions of the mock image in xy coordinates. If int is given,
        will make the x and y dimensions the same.
    pixel_scale : float or `~astropy.units.Quantity`
        The pixel scale of the mock image. If a float is given, the units will
        be assumed to be `arcsec / pixels`. Default is `0.2 arcsec / pixel`.
    sb : float
        Surface brightness of of uniform stellar distribution.
    sb_band : str
        Photometric filter of ``sb``. Must be part of ``phot_system``.
    mag_limit : float, optional
        Only sample individual stars that are brighter than this magnitude. All
        fainter stars will be combined into an integrated component. Otherwise,
        all stars in the population will be sampled. You must also give the
        `mag_limit_band` if you use this parameter.
    mag_limit_band : str, optional
        Bandpass of the limiting magnitude. You must give this parameter if
        you use the `mag_limit` parameter.
    imf : str, optional
        The initial stellar mass function. Default is `'kroupa'`.
    imf_kw : dict, optional
        Optional keyword arguments for sampling the stellar mass function.
    ssp_kw : dict, optional
        Additional keyword arguments for `~artpop.stars.SSP`.
    mist_path : str, optional
        Path to MIST isochrone grids. Use this if you want to use a different
        path from the `MIST_PATH` environment variable.
    label_type : str, optional
        If not None, the type of labels to pass to the `~artpop.source.Source`
        object. Can be `phases` or an attribute of `~artpop.stars.MISTSSP`.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    """

    def __init__(self, log_age, feh, phot_system, distance, xy_dim,
                 pixel_scale, sb, sb_band, mag_limit=None, mag_limit_band=None,
                 imf='kroupa', imf_kw={}, ssp_kw={}, mist_path=MIST_PATH,
                 label_type=None, random_state=None, **kwargs):

        self.iso_kw = dict(log_age=log_age, feh=feh, phot_system=phot_system,
                           mist_path=mist_path)
        self.iso_kw.update(kwargs)

        iso = MISTIsochrone(**self.iso_kw)

        super(MISTUniformSSP, self).__init__(
            isochrone=iso, distance=distance, xy_dim=xy_dim, sb=sb,
            sb_band=sb_band, mag_limit=mag_limit,
            mag_limit_band=mag_limit_band, pixel_scale=pixel_scale, imf=imf,
            imf_kw=imf_kw, ssp_kw=ssp_kw, label_type=label_type,
            random_state=random_state)



# Third-party
import numpy as np
from scipy.integrate import quad
from scipy.interpolate import interp1d
from astropy import units as u

# Project
from ..util import check_random_state, check_units


__all__ = ['IMFIntegrator', 'broken_power_law', 'broken_power_law_normed',
           'salpeter', 'kroupa', 'scalo', 'imf_dict', 'sample_imf',
           'build_galaxy']


def salpeter(mass_grid, norm_type=None,
             norm_mass_min=None, norm_mass_max=None):
    """
    Calculate weights for the Salpeter IMF (`Salpeter 1955
    <https://ui.adsabs.harvard.edu/abs/1955ApJ...121..161S/abstract>`_).

    Parameters
    ----------
    mass_grid : `~numpy.ndarray`
        Stellar mass grid.
    norm_type : str, optional
        How to normalize the weights: by 'number', 'mass', or the 'sum'.
    norm_mass_min : int or None, optional
        Minimum mass to use for normalization. If None, use minimum of
        `mass_grid` will be used.
    norm_mass_max : int or None, optional
        Maximum mass to use for normalization. If None, use maximum of
        `mass_grid` will be used.

    Returns
    -------
    weights : `~numpy.ndarray`
        The weights associated with each mass in the input `mass_grid`.
    """
    alpha = 2.35

    weights =  mass_grid**-alpha

    if norm_type is None:
        return weights

    m_min = norm_mass_min if norm_mass_min else mass_grid.min()
    m_max = norm_mass_max if norm_mass_max else mass_grid.max()

    if norm_type == 'number':
        beta = alpha - 1
        norm = beta / (m_min**-beta - m_max**-beta)
    elif norm_type == 'mass':
        beta = alpha - 2
        norm = beta / (m_min**-beta - m_max**-beta)
    elif norm_type == 'sum':
        norm = 1 / weights.sum()

    weights *= norm

    return weights


def broken_power_law(mass_grid, alphas, breaks):
    """
    Calculate weights for broken power law.

    Parameters
    ----------
    mass_grid : `~numpy.ndarray`
        Stellar mass grid.
    alphas : list-like
        The three slopes of the broken power law.
    breaks  : list-like
        The two break point masses of the broken power law.

    Returns
    -------
    weights : `~numpy.ndarray`
        The weights associated with each mass in the input `mass_grid`.
    """
    mass_grid = np.asarray(mass_grid)
    a1, a2, a3 = alphas
    b1, b2 = breaks
    alpha = np.where(mass_grid < b1, a1, np.where(mass_grid < b2, a2, a3))
    m_break = np.where(mass_grid < b2, b1, b2 * (b1 / b2)**(a2 / a3))
    weights = (mass_grid / m_break)**(-alpha)
    return weights


def broken_power_law_normed(mass_grid, alphas, breaks, norm_type='number',
                            num_norm_bins=1e5, norm_mass_min=None,
                            norm_mass_max=None):
    """
    Calculate normalized weights for broken power law.

    Parameters
    ----------
    mass_grid : `~numpy.ndarray`
        Stellar mass grid.
    alphas : list-like
        The three slopes of the broken power law.
    breaks  : list-like
        The two break point masses of the broken power law.
    norm_type : str, optional
        How to normalize the weights: by 'number', 'mass', or the 'sum'.
    num_norm_bins : int, optional
        Number of mass bins to use for integration (if needed to normalize).
    norm_mass_min : int or None, optional
        Minimum mass to use for normalization. If None, use minimum of
        `mass_grid` will be used.
    norm_mass_max : int or None, optional
        Maximum mass to use for normalization. If None, use maximum of
        `mass_grid` will be used.

    Returns
    -------
    weights : `~numpy.ndarray`
        The weights associated with each mass in the input `mass_grid`.
    """
    weights = broken_power_law(mass_grid, alphas, breaks)
    if norm_type ==  'sum':
        norm = weights.sum()
    else:
        m_min = norm_mass_min if norm_mass_min else mass_grid.min()
        m_max = norm_mass_max if norm_mass_max else mass_grid.max()
        m_integrate = np.linspace(m_min, m_max, int(num_norm_bins))
        w_integrate = broken_power_law(m_integrate, alphas, breaks)
        if norm_type == 'number':
            norm = np.trapz(w_integrate, m_integrate)
        elif norm_type == 'mass':
            w_integrate = m_integrate * w_integrate
            norm = np.trapz(w_integrate, m_integrate)
    weights /= norm
    return weights


def kroupa(mass_grid, norm_type=None, num_norm_bins=1e5,
           norm_mass_min=None, norm_mass_max=None):
    """
    The Kroupa stellar initial mass function (`Kroupa 2001
    <https://ui.adsabs.harvard.edu/abs/2001MNRAS.322..231K/abstract>`_).

    Parameters
    ----------
    mass_grid : `~numpy.ndarray`
        Stellar mass grid.
    norm_type : str, optional
        How to normalize the weights: by 'number', 'mass', or the 'sum'.
    num_norm_bins : int, optional
        Number of mass bins to use for integration (if needed to normalize).
    norm_mass_min : int or None, optional
        Minimum mass to use for normalization. If None, use minimum of
        `mass_grid` will be used.
    norm_mass_max : int or None, optional
        Maximum mass to use for normalization. If None, use maximum of
        `mass_grid` will be used.

    Returns
    -------
    weights : `~numpy.ndarray`
        The weights associated with each mass in the input `mass_grid`.
    """
    alphas = [0.3, 1.3, 2.3]
    breaks = [0.08, 0.5]
    if norm_type is not None:
        kw = dict(norm_type=norm_type, num_norm_bins=num_norm_bins,
                  norm_mass_min=norm_mass_min, norm_mass_max=norm_mass_max)
        weights = broken_power_law_normed(mass_grid, alphas, breaks, **kw)
    else:
        weights = broken_power_law(mass_grid, alphas, breaks)
    return weights


def scalo(mass_grid, norm_type=None, num_norm_bins=1e5,
          norm_mass_min=None, norm_mass_max=None):
    """
    The Scalo stellar initial mass function (`Scalo 1998
    <https://ui.adsabs.harvard.edu/abs/1998ASPC..142..201S/abstract>`_).

    Parameters
    ----------
    mass_grid : `~numpy.ndarray`
        Stellar mass grid.
    norm_type : str, optional
        How to normalize the weights: by 'number', 'mass', or the 'sum'.
    num_norm_bins : int, optional
        Number of mass bins to use for integration (if needed to normalize).
    norm_mass_min : int or None, optional
        Minimum mass to use for normalization. If None, use minimum of
        `mass_grid` will be used.
    norm_mass_max : int or None, optional
        Maximum mass to use for normalization. If None, use maximum of
        `mass_grid` will be used.

    Returns
    -------
    weights : `~numpy.ndarray`
        The weights associated with each mass in the input `mass_grid`.
    """
    alphas = [1.2, 2.7, 2.3]
    breaks = [1, 10]
    if norm_type is not None:
        kw = dict(norm_type=norm_type, num_norm_bins=num_norm_bins,
                  norm_mass_min=norm_mass_min, norm_mass_max=norm_mass_max)
        weights = broken_power_law_normed(mass_grid, alphas, breaks, **kw)
    else:
        weights = broken_power_law(mass_grid, alphas, breaks)
    return weights


# IMF helper dictionary
imf_dict = dict(salpeter=salpeter, kroupa=kroupa, scalo=scalo)


def sample_imf(num_stars, m_min=0.08, m_max=120, imf='kroupa',
               num_mass_bins=100000, random_state=None, imf_kw={}):
    """
    Sample stellar IMF via inverse transform sampling.

    Parameters
    ----------
    num_stars : int
        Number of stars to sample.
    m_min : float, optional
        Minimum stellar mass.
    m_max : float, optional
        Maximum stellar mass.
    imf : str, optional
        The desired IMF (salpeter or kroupa)
    num_mass_bins : int, optional
        Number of mass bins in logarithmic spaced mass grid.
    random_state : `None`, int, list of ints, or `~numpy.random.RandomState`
        If `None`, return the `~numpy.random.RandomState` singleton used by
        ``numpy.random``. If `int`, return a new `~numpy.random.RandomState`
        instance seeded with the `int`.  If `~numpy.random.RandomState`,
        return it. Otherwise raise ``ValueError``.
    imf_kw : dict, optional
        Keyword arguments for the imf function.

    Returns
    -------
    masses : `~numpy.ndarray`
        The sampled stellar masses.
    """
    rng = check_random_state(random_state)
    bin_edges = np.logspace(np.log10(m_min),
                            np.log10(m_max),
                            int(num_mass_bins))
    mass_grid = (bin_edges[1:] + bin_edges[:-1]) / 2.0
    weights = imf_dict[imf](mass_grid, **imf_kw)
    dm = np.diff(bin_edges)
    cdf = np.cumsum(weights * dm)
    cdf /= cdf.max()
    cdf = interp1d(cdf, mass_grid, bounds_error=False, fill_value=m_min)
    rand_num = rng.uniform(low=0., high=1.0, size=int(num_stars))
    masses = cdf(rand_num)
    return masses


def build_galaxy(stellar_mass, num_stars_iter=1e5, **kwargs):
    """
    Build galaxy of a given stellar mass.

    Parameters
    ----------
    stellar_mass : float or `~astropy.units.Quantity`
        Stellar mass of galaxy. If float is given, the units are assumed to be
        solar masses.
    num_stars_iter : int
        Number of stars to generate at each iteration. Lower this
        number (at the expense of speed) to get a more accurate total mass.

    Returns
    -------
    stars : `~numpy.ndarray`
        Stellar masses of all the stars.
    """

    stars = []
    total_mass = 0.0
    stellar_mass = check_units(stellar_mass, 'Msun').to('Msun').value

    while total_mass < stellar_mass:
        new_stars = sample_imf(int(num_stars_iter), **kwargs)
        total_mass += new_stars.sum()
        stars = np.concatenate([stars, new_stars])

    return stars


class IMFIntegrator(object):
    """
    A helper class for numerically integrating the IMF.
    """

    def __init__(self, name, m_min=0.1, m_max=120.0):
        self.name = name
        self._func = imf_dict[name]
        self.m_min = m_min
        self.m_max = m_max
        self.num_norm = self.integrate(m_min, m_max, None)
        self.mass_norm = self.m_integrate(m_min, m_max, None)

    def func(self, m):
        return self._func(m, norm_type=None)

    def m_func(self, m):
        return m * self._func(m, norm_type=None)

    def integrate(self, m_min=None, m_max=None, norm=False, **kwargs):
        m_min = m_min if m_min else self.m_min
        m_max = m_max if m_max else self.m_max
        if norm == True:
            n = self.num_norm
        elif norm is None or norm == False:
            n = 1.0
        elif type(norm) == int or type(norm) == float:
            n = norm
        else:
            raise Exception(f'{norm} is not a valid normalization.')
        return quad(self.func, m_min, m_max, **kwargs)[0] / n

    def m_integrate(self, m_min=None, m_max=None, norm=False, **kwargs):
        m_min = m_min if m_min else self.m_min
        m_max = m_max if m_max else self.m_max
        if norm == True:
            n = self.mass_norm
        elif norm is None or norm == False:
            n = 1.0
        elif type(norm) == int or type(norm) == float:
            n = norm
        else:
            raise Exception(f'{norm} is not a valid normalization.')
        return quad(self.m_func, m_min, m_max, **kwargs)[0] / n



# Third-party
import numpy as np
import matplotlib.pyplot as plt


__all__ = ['mpl_style', 'jpg_style', 'show_image']


mpl_style = {
    'axes.spines.left': True,
    'axes.spines.bottom': True,
    'axes.spines.top': True,
    'axes.spines.right': True,
    'axes.edgecolor': 'black',
    'axes.linewidth': 2.0,
    'axes.titlesize': 'xx-large',
    'axes.labelsize': 'xx-large',

    'xtick.major.size': 8,
    'xtick.minor.size': 4,
    'xtick.major.width': 2.0,
    'xtick.minor.width': 2.0,
    'xtick.direction': 'in',
    'xtick.labelsize': 'x-large',

    'ytick.major.size': 8,
    'ytick.minor.size': 4,
    'ytick.major.width': 2.0,
    'ytick.minor.width': 2.0,
    'ytick.direction': 'in',
    'ytick.labelsize': 'x-large',

    'xtick.top': True,
    'ytick.right': True,

    'legend.numpoints': 1,
    'legend.fontsize': 'x-large',
    'legend.handletextpad': 0.3,
    'legend.frameon': False,
    'legend.scatterpoints': 1,
    'savefig.bbox': 'tight'
}


jpg_style = {
    'font.family': 'serif',
    'font.serif': 'Times New Roman',

    'axes.spines.left': True,
    'axes.spines.bottom': True,
    'axes.spines.top': True,
    'axes.spines.right' : True,
    'axes.edgecolor': 'black',
    'axes.linewidth': 2.0,
    'axes.titlesize': 18,
    'axes.labelsize': 18,

    'xtick.major.size': 10,
    'xtick.minor.size': 5,
    'xtick.major.width': 2.0,
    'xtick.minor.width': 2.0,
    'xtick.labelsize': 15,
    'xtick.direction': 'in',
    'ytick.major.size': 10,
    'ytick.minor.size': 5,
    'ytick.major.width': 2.0,
    'ytick.minor.width': 2.0,
    'ytick.labelsize': 15,
    'ytick.direction': 'in',
    'xtick.top': True,
    'ytick.right': True,

    'legend.numpoints': 1,
    'legend.handletextpad': 0.3,
    'legend.frameon': False,
    'legend.scatterpoints' : 1,

    'savefig.format': 'pdf',
    'savefig.bbox': 'tight',

    'text.usetex': True
}


def show_image(image, percentile=[0.1, 99.9], subplots=None, cmap='gray_r',
               figsize=(10, 10), **kwargs):
    """
    Display image using matplotlib.

    Parameters
    ----------
    image : `~numpy.ndarray`
        The image pixels.
    percentile : list-like or None, optional
        Set the min and max pixel values to the given low and high percentile
        values: [low, high]. If None, use all pixel values.
    subplots : tuple or None, optional
        The ``matplotlib`` figure and axis objects (`fig`, `ax`). If None, a
        new figure will be created.
    cmap : str, optional
        ``matplotlib`` color map.
    figsize : tuple, optional
         Figure size. Only used if subplots is None.
    **kwargs
        Keyword arguments for `~matplotlib.pyplot.imshow`.

    Returns
    -------
    fig : `~matplotlib.figure.Figure`
        The Figure object.
    ax : `~matplotlib.axes._subplots.AxesSubplot`
        The axis.
    """
    if subplots is None:
        fig, ax = plt.subplots(figsize=figsize,
                               subplot_kw=dict(xticks=[], yticks=[]))

    else:
        fig, ax = subplots
    if percentile is not None:
        vmin, vmax = np.nanpercentile(image, percentile)
    else:
        vmin, vmax = None, None

    interp = kwargs.pop('interpolation', None)
    ax.imshow(image, origin='lower', cmap=cmap, vmin=vmin, vmax=vmax,
              interpolation=interp, **kwargs)

    return fig, ax
